import os
import sys
import tempfile

try:
    DATA_PATH = os.environ["ARCOR2_DATA_PATH"]
except KeyError:
    sys.exit("'ARCOR2_DATA_PATH' env. variable not set.")

URDF_PATH = os.path.join(DATA_PATH, "urdf")

OBJECT_TYPE_PATH = tempfile.mkdtemp()
OBJECT_TYPE_MODULE = "arcor2_object_types"



import asyncio
import copy
from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator, Dict, List, Optional, Set

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import helpers as hlp
from arcor2 import transformations as tr
from arcor2.cached import CachedProjectException, CachedScene, UpdateableCachedProject
from arcor2.data import common
from arcor2.data.events import Event, PackageState
from arcor2.exceptions import Arcor2Exception
from arcor2.logic import LogicContainer, check_for_loops
from arcor2.object_types.abstract import Robot
from arcor2.parameter_plugins.base import ParameterPluginException
from arcor2.parameter_plugins.pose import PosePlugin
from arcor2.parameter_plugins.utils import plugin_from_type_name
from arcor2_arserver import globals as glob
from arcor2_arserver import notifications as notif
from arcor2_arserver import project
from arcor2_arserver.clients import persistent_storage as storage
from arcor2_arserver.decorators import no_project, project_needed, scene_needed
from arcor2_arserver.helpers import unique_name
from arcor2_arserver.objects_actions import get_types_dict
from arcor2_arserver.project import (
    check_action_params,
    check_flows,
    close_project,
    find_object_action,
    open_project,
    project_names,
    project_problems,
)
from arcor2_arserver.robot import get_end_effector_pose, get_robot_joints
from arcor2_arserver.scene import can_modify_scene, ensure_scene_started, get_instance, open_scene
from arcor2_arserver_data import events as sevts
from arcor2_arserver_data import rpc as srpc


@asynccontextmanager
async def managed_project(project_id: str, make_copy: bool = False) -> AsyncGenerator[UpdateableCachedProject, None]:

    save_back = False

    if glob.PROJECT and glob.PROJECT.id == project_id:
        if make_copy:
            project = copy.deepcopy(glob.PROJECT)
            save_back = True
        else:
            project = glob.PROJECT
    else:
        save_back = True
        project = UpdateableCachedProject(await storage.get_project(project_id))

    if make_copy:
        project.id = common.Project.uid()

    try:
        yield project
    finally:
        if save_back:
            asyncio.ensure_future(storage.update_project(project.project))


@scene_needed
@project_needed
async def cancel_action_cb(req: srpc.p.CancelAction.Request, ui: WsClient) -> None:

    assert glob.SCENE
    assert glob.PROJECT

    if not glob.RUNNING_ACTION:
        raise Arcor2Exception("No action is running.")

    action = glob.PROJECT.action(glob.RUNNING_ACTION)
    obj_id, action_name = action.parse_type()
    obj = get_instance(obj_id)

    if not find_object_action(glob.SCENE, action).meta.cancellable:
        raise Arcor2Exception("Action is not cancellable.")

    try:
        action_method = getattr(obj, action_name)
        cancel_method = getattr(obj, obj.CANCEL_MAPPING[action_method.__name__])
    except AttributeError as e:
        raise Arcor2Exception("Internal error.") from e

    cancel_params: Dict[str, Any] = {}

    # TODO fix or remove this?
    """
    cancel_sig = inspect.signature(cancel_method)

    assert glob.RUNNING_ACTION_PARAMS is not None

    for param_name in cancel_sig.parameters.keys():
        try:
            cancel_params[param_name] = glob.RUNNING_ACTION_PARAMS[param_name]
        except KeyError as e:
            raise Arcor2Exception("Cancel method parameters should be subset of action parameters.") from e
    """

    await hlp.run_in_executor(cancel_method, *cancel_params.values())

    asyncio.ensure_future(notif.broadcast_event(sevts.a.ActionCancelled()))
    glob.RUNNING_ACTION = None
    glob.RUNNING_ACTION_PARAMS = None


@scene_needed
@project_needed
async def execute_action_cb(req: srpc.p.ExecuteAction.Request, ui: WsClient) -> None:

    assert glob.SCENE
    assert glob.PROJECT

    ensure_scene_started()

    if glob.RUNNING_ACTION:
        raise Arcor2Exception(
            f"Action {glob.RUNNING_ACTION} is being executed. " f"Only one action can be executed at a time."
        )

    action = glob.PROJECT.action(req.args.action_id)

    obj_id, action_name = action.parse_type()

    params: List[Any] = []

    for param in action.parameters:

        if param.type == common.ActionParameter.TypeEnum.LINK:

            parsed_link = param.parse_link()
            try:
                results = project.PREV_RESULTS[parsed_link.action_id]
            except KeyError:
                prev_action = glob.PROJECT.action(parsed_link.action_id)
                raise Arcor2Exception(f"Action '{prev_action.name}' has to be executed first.")

            # an action result could be a tuple or a single value
            if isinstance(results, tuple):
                params.append(results[parsed_link.output_index])
            else:
                assert parsed_link.output_index == 0
                params.append(results)

        elif param.type == common.ActionParameter.TypeEnum.CONSTANT:
            const = glob.PROJECT.constant(param.str_from_value())
            # TODO use plugin to get the value
            import json

            params.append(json.loads(const.value))
        else:

            try:
                params.append(
                    plugin_from_type_name(param.type).parameter_execution_value(
                        get_types_dict(), glob.SCENE, glob.PROJECT, action.id, param.name
                    )
                )
            except ParameterPluginException as e:
                glob.logger.error(e)
                raise Arcor2Exception(f"Failed to get value for parameter {param.name}.")

    obj = get_instance(obj_id)

    if isinstance(obj, Robot):
        if obj.move_in_progress:
            raise Arcor2Exception("Can't execute actions while the robot moves.")

    if not hasattr(obj, action_name):
        raise Arcor2Exception("Internal error: object does not have the requested method.")

    if req.dry_run:
        return

    glob.RUNNING_ACTION = action.id
    glob.RUNNING_ACTION_PARAMS = params

    glob.logger.debug(f"Running action {action.name} ({type(obj)}/{action_name}), params: {params}.")

    # schedule execution and return success
    asyncio.ensure_future(project.execute_action(getattr(obj, action_name), params))
    return None


async def project_info(
    project_id: str, scenes_lock: asyncio.Lock, scenes: Dict[str, CachedScene]
) -> srpc.p.ListProjects.Response.Data:

    project = await storage.get_project(project_id)

    assert project.modified is not None

    pd = srpc.p.ListProjects.Response.Data(
        id=project.id, desc=project.desc, name=project.name, scene_id=project.scene_id, modified=project.modified
    )

    try:
        cached_project = UpdateableCachedProject(project)
    except CachedProjectException as e:
        pd.problems.append(str(e))
        return pd

    try:
        async with scenes_lock:
            if project.scene_id not in scenes:
                scenes[project.scene_id] = CachedScene(await storage.get_scene(project.scene_id))
    except storage.ProjectServiceException:
        pd.problems.append("Scene does not exist.")
        return pd

    pd.problems = project_problems(scenes[project.scene_id], cached_project)
    pd.valid = not pd.problems

    if not pd.valid:
        return pd

    try:
        # TODO call build service!!!
        pd.executable = True
    except Arcor2Exception as e:
        pd.problems.append(str(e))

    return pd


async def list_projects_cb(req: srpc.p.ListProjects.Request, ui: WsClient) -> srpc.p.ListProjects.Response:

    projects = await storage.get_projects()

    scenes_lock = asyncio.Lock()
    scenes: Dict[str, CachedScene] = {}

    resp = srpc.p.ListProjects.Response()
    tasks = [project_info(project_iddesc.id, scenes_lock, scenes) for project_iddesc in projects.items]

    resp.data = []
    for res in await asyncio.gather(*tasks, return_exceptions=True):

        if isinstance(res, Arcor2Exception):
            glob.logger.error(str(res))
        elif isinstance(res, Exception):
            raise res  # zero toleration for other exceptions
        else:
            resp.data.append(res)

    return resp


@scene_needed
@project_needed
async def add_action_point_joints_using_robot_cb(
    req: srpc.p.AddActionPointJointsUsingRobot.Request, ui: WsClient
) -> None:

    ensure_scene_started()

    assert glob.SCENE
    assert glob.PROJECT

    ap = glob.PROJECT.bare_action_point(req.args.action_point_id)

    hlp.is_valid_identifier(req.args.name)
    unique_name(req.args.name, glob.PROJECT.ap_joint_names(ap.id))

    new_joints = await get_robot_joints(req.args.robot_id)

    if req.dry_run:
        return None

    prj = common.ProjectRobotJoints(req.args.name, req.args.robot_id, new_joints, True)
    glob.PROJECT.upsert_joints(ap.id, prj)

    evt = sevts.p.JointsChanged(prj)
    evt.change_type = Event.Type.ADD
    evt.parent_id = ap.id
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


@scene_needed
@project_needed
async def update_action_point_joints_using_robot_cb(
    req: srpc.p.UpdateActionPointJointsUsingRobot.Request, ui: WsClient
) -> None:

    ensure_scene_started()

    assert glob.SCENE
    assert glob.PROJECT

    robot_joints = glob.PROJECT.joints(req.args.joints_id)
    robot_joints.joints = await get_robot_joints(robot_joints.robot_id)
    robot_joints.is_valid = True

    glob.PROJECT.update_modified()

    evt = sevts.p.JointsChanged(robot_joints)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


@scene_needed
@project_needed
async def update_action_point_joints_cb(req: srpc.p.UpdateActionPointJoints.Request, ui: WsClient) -> None:

    assert glob.SCENE
    assert glob.PROJECT

    robot_joints = glob.PROJECT.joints(req.args.joints_id)

    if {joint.name for joint in req.args.joints} != {joint.name for joint in robot_joints.joints}:
        raise Arcor2Exception("Joint names does not match the robot.")

    # TODO maybe joints values should be normalized? To <0, 2pi> or to <-pi, pi>?
    robot_joints.joints = req.args.joints
    robot_joints.is_valid = True
    glob.PROJECT.update_modified()

    evt = sevts.p.JointsChanged(robot_joints)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


@scene_needed
@project_needed
async def remove_action_point_joints_cb(req: srpc.p.RemoveActionPointJoints.Request, ui: WsClient) -> None:
    """Removes joints from action point.

    :param req:
    :return:
    """

    assert glob.SCENE
    assert glob.PROJECT

    for act in glob.PROJECT.actions:
        for param in act.parameters:
            if plugin_from_type_name(param.type).uses_robot_joints(
                glob.PROJECT, act.id, param.name, req.args.joints_id
            ):
                raise Arcor2Exception(f"Joints used in action {act.name} (parameter {param.name}).")

    joints_to_be_removed = glob.PROJECT.remove_joints(req.args.joints_id)

    glob.PROJECT.update_modified()

    evt = sevts.p.JointsChanged(joints_to_be_removed)
    evt.change_type = Event.Type.REMOVE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


@scene_needed
@project_needed
async def rename_action_point_cb(req: srpc.p.RenameActionPoint.Request, ui: WsClient) -> None:

    assert glob.SCENE
    assert glob.PROJECT

    ap = glob.PROJECT.bare_action_point(req.args.action_point_id)

    if req.args.new_name == ap.name:
        return None

    hlp.is_valid_identifier(req.args.new_name)
    unique_name(req.args.new_name, glob.PROJECT.action_points_names)

    if req.dry_run:
        return None

    ap.name = req.args.new_name

    glob.PROJECT.update_modified()

    evt = sevts.p.ActionPointChanged(ap)
    evt.change_type = Event.Type.UPDATE_BASE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


def detect_ap_loop(ap: common.BareActionPoint, new_parent_id: str) -> None:

    assert glob.PROJECT

    visited_ids: Set[str] = set()
    ap = copy.deepcopy(ap)
    ap.parent = new_parent_id
    while True:

        if ap.id in visited_ids:
            raise Arcor2Exception("Loop detected!")

        visited_ids.add(ap.id)

        if ap.parent is None:
            break  # type: ignore  # this is certainly reachable

        try:
            ap = glob.PROJECT.bare_action_point(ap.parent)
        except Arcor2Exception:
            break


@scene_needed
@project_needed
async def update_action_point_parent_cb(req: srpc.p.UpdateActionPointParent.Request, ui: WsClient) -> None:

    assert glob.SCENE
    assert glob.PROJECT

    ap = glob.PROJECT.bare_action_point(req.args.action_point_id)

    if req.args.new_parent_id == ap.parent:
        return None

    check_ap_parent(req.args.new_parent_id)
    detect_ap_loop(ap, req.args.new_parent_id)

    if req.dry_run:
        return

    if not ap.parent and req.args.new_parent_id:
        # AP position and all orientations will become relative to the parent
        tr.make_global_ap_relative(glob.SCENE, glob.PROJECT, ap, req.args.new_parent_id)

    elif ap.parent and not req.args.new_parent_id:
        # AP position and all orientations will become absolute
        tr.make_relative_ap_global(glob.SCENE, glob.PROJECT, ap)
    else:

        assert ap.parent

        # AP position and all orientations will become relative to another parent
        tr.make_relative_ap_global(glob.SCENE, glob.PROJECT, ap)
        tr.make_global_ap_relative(glob.SCENE, glob.PROJECT, ap, req.args.new_parent_id)

    ap.parent = req.args.new_parent_id
    glob.PROJECT.update_modified()

    """
    Can't send orientation changes and then ActionPointChanged/UPDATE_BASE (or vice versa)
    because UI would display orientations wrongly (for a short moment).
    """
    # 'ap' is BareActionPoint, that does not contain orientations
    evt = sevts.p.ActionPointChanged(glob.PROJECT.action_point(req.args.action_point_id))
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def update_ap_position(ap: common.BareActionPoint, position: common.Position) -> None:
    """Updates position of an AP and sends notification about joints that
    become invalid because of it.

    :param ap_id:
    :param position:
    :return:
    """

    assert glob.PROJECT

    valid_joints = [joints for joints in glob.PROJECT.ap_joints(ap.id) if joints.is_valid]
    glob.PROJECT.update_ap_position(ap.id, position)

    for joints in valid_joints:  # those are now invalid, so let's notify UI about the change

        assert not joints.is_valid

        evt = sevts.p.JointsChanged(joints)
        evt.change_type = Event.Type.UPDATE
        evt.parent_id = ap.id
        asyncio.ensure_future(notif.broadcast_event(evt))

    ap_evt = sevts.p.ActionPointChanged(ap)
    ap_evt.change_type = Event.Type.UPDATE_BASE
    asyncio.ensure_future(notif.broadcast_event(ap_evt))


@scene_needed
@project_needed
async def update_action_point_position_cb(req: srpc.p.UpdateActionPointPosition.Request, ui: WsClient) -> None:

    assert glob.PROJECT

    ap = glob.PROJECT.bare_action_point(req.args.action_point_id)

    if req.dry_run:
        return

    await update_ap_position(ap, req.args.new_position)


@scene_needed
@project_needed
async def update_action_point_using_robot_cb(req: srpc.p.UpdateActionPointUsingRobot.Request, ui: WsClient) -> None:

    ensure_scene_started()

    assert glob.SCENE
    assert glob.PROJECT

    ap = glob.PROJECT.bare_action_point(req.args.action_point_id)
    new_pose = await get_end_effector_pose(req.args.robot.robot_id, req.args.robot.end_effector)

    if ap.parent:
        new_pose = tr.make_pose_rel_to_parent(glob.SCENE, glob.PROJECT, new_pose, ap.parent)

    await update_ap_position(glob.PROJECT.bare_action_point(req.args.action_point_id), new_pose.position)
    return None


@scene_needed
@project_needed
async def add_action_point_orientation_cb(req: srpc.p.AddActionPointOrientation.Request, ui: WsClient) -> None:
    """Adds orientation and joints to the action point.

    :param req:
    :return:
    """

    assert glob.SCENE
    assert glob.PROJECT

    ap = glob.PROJECT.bare_action_point(req.args.action_point_id)
    hlp.is_valid_identifier(req.args.name)
    unique_name(req.args.name, glob.PROJECT.ap_orientation_names(ap.id))

    if req.dry_run:
        return None

    orientation = common.NamedOrientation(req.args.name, req.args.orientation)
    glob.PROJECT.upsert_orientation(ap.id, orientation)

    evt = sevts.p.OrientationChanged(orientation)
    evt.change_type = Event.Type.ADD
    evt.parent_id = ap.id
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


@scene_needed
@project_needed
async def update_action_point_orientation_cb(req: srpc.p.UpdateActionPointOrientation.Request, ui: WsClient) -> None:
    """Updates orientation of the action point.

    :param req:
    :return:
    """

    assert glob.SCENE
    assert glob.PROJECT

    orientation = glob.PROJECT.orientation(req.args.orientation_id)
    orientation.orientation = req.args.orientation

    glob.PROJECT.update_modified()

    evt = sevts.p.OrientationChanged(orientation)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


@scene_needed
@project_needed
async def add_action_point_orientation_using_robot_cb(
    req: srpc.p.AddActionPointOrientationUsingRobot.Request, ui: WsClient
) -> None:
    """Adds orientation and joints to the action point.

    :param req:
    :return:
    """

    ensure_scene_started()

    assert glob.SCENE
    assert glob.PROJECT

    ap = glob.PROJECT.bare_action_point(req.args.action_point_id)
    hlp.is_valid_identifier(req.args.name)
    unique_name(req.args.name, glob.PROJECT.ap_orientation_names(ap.id))

    if req.dry_run:
        return None

    new_pose = await get_end_effector_pose(req.args.robot.robot_id, req.args.robot.end_effector)

    if ap.parent:
        new_pose = tr.make_pose_rel_to_parent(glob.SCENE, glob.PROJECT, new_pose, ap.parent)

    orientation = common.NamedOrientation(req.args.name, new_pose.orientation)
    glob.PROJECT.upsert_orientation(ap.id, orientation)

    evt = sevts.p.OrientationChanged(orientation)
    evt.change_type = Event.Type.ADD
    evt.parent_id = ap.id
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


@scene_needed
@project_needed
async def update_action_point_orientation_using_robot_cb(
    req: srpc.p.UpdateActionPointOrientationUsingRobot.Request, ui: WsClient
) -> None:
    """Updates orientation and joint of the action point.

    :param req:
    :return:
    """

    ensure_scene_started()

    assert glob.SCENE
    assert glob.PROJECT

    ap, ori = glob.PROJECT.bare_ap_and_orientation(req.args.orientation_id)

    new_pose = await get_end_effector_pose(req.args.robot.robot_id, req.args.robot.end_effector)

    if ap.parent:
        new_pose = tr.make_pose_rel_to_parent(glob.SCENE, glob.PROJECT, new_pose, ap.parent)

    ori.orientation = new_pose.orientation

    glob.PROJECT.update_modified()

    evt = sevts.p.OrientationChanged(ori)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


@scene_needed
@project_needed
async def remove_action_point_orientation_cb(req: srpc.p.RemoveActionPointOrientation.Request, ui: WsClient) -> None:
    """Removes orientation.

    :param req:
    :return:
    """

    assert glob.SCENE
    assert glob.PROJECT

    ap, orientation = glob.PROJECT.bare_ap_and_orientation(req.args.orientation_id)

    for act in glob.PROJECT.actions:
        for param in act.parameters:
            if plugin_from_type_name(param.type).uses_orientation(
                glob.PROJECT, act.id, param.name, req.args.orientation_id
            ):
                raise Arcor2Exception(f"Orientation used in action {act.name} (parameter {param.name}).")

    if req.dry_run:
        return None

    glob.PROJECT.remove_orientation(req.args.orientation_id)

    evt = sevts.p.OrientationChanged(orientation)
    evt.change_type = Event.Type.REMOVE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def open_project_cb(req: srpc.p.OpenProject.Request, ui: WsClient) -> None:

    if glob.PACKAGE_STATE.state in PackageState.RUN_STATES:
        raise Arcor2Exception("Can't open project while package runs.")

    # TODO validate using project_problems?
    await open_project(req.args.id)

    assert glob.SCENE
    assert glob.PROJECT

    asyncio.ensure_future(
        notif.broadcast_event(sevts.p.OpenProject(sevts.p.OpenProject.Data(glob.SCENE.scene, glob.PROJECT.project)))
    )

    return None


@scene_needed
@project_needed
async def save_project_cb(req: srpc.p.SaveProject.Request, ui: WsClient) -> None:

    assert glob.SCENE
    assert glob.PROJECT

    glob.PROJECT.modified = await storage.update_project(glob.PROJECT.project)
    asyncio.ensure_future(notif.broadcast_event(sevts.p.ProjectSaved()))
    return None


@no_project
async def new_project_cb(req: srpc.p.NewProject.Request, ui: WsClient) -> None:

    if glob.PACKAGE_STATE.state in PackageState.RUN_STATES:
        raise Arcor2Exception("Can't create project while package runs.")

    unique_name(req.args.name, (await project_names()))

    if req.dry_run:
        return None

    if glob.SCENE:
        if glob.SCENE.id != req.args.scene_id:
            raise Arcor2Exception("Another scene is opened.")

        if glob.SCENE.has_changes():
            glob.SCENE.modified = await storage.update_scene(glob.SCENE.scene)
    else:

        if req.args.scene_id not in {scene.id for scene in (await storage.get_scenes()).items}:
            raise Arcor2Exception("Unknown scene id.")

        await open_scene(req.args.scene_id)

    project.PREV_RESULTS.clear()
    glob.PROJECT = UpdateableCachedProject(
        common.Project(req.args.name, req.args.scene_id, desc=req.args.desc, has_logic=req.args.has_logic)
    )

    assert glob.SCENE

    asyncio.ensure_future(
        notif.broadcast_event(sevts.p.OpenProject(sevts.p.OpenProject.Data(glob.SCENE.scene, glob.PROJECT.project)))
    )
    return None


@scene_needed
@project_needed
async def close_project_cb(req: srpc.p.CloseProject.Request, ui: WsClient) -> None:

    can_modify_scene()

    assert glob.PROJECT

    if not req.args.force and glob.PROJECT.has_changes:
        raise Arcor2Exception("Project has unsaved changes.")

    if req.dry_run:
        return None

    await close_project()
    return None


def check_ap_parent(parent: Optional[str]) -> None:

    assert glob.SCENE
    assert glob.PROJECT

    if not parent:
        return

    if parent in glob.SCENE.object_ids:
        if glob.SCENE.object(parent).pose is None:
            raise Arcor2Exception("AP can't have object without pose as parent.")
    elif parent not in glob.PROJECT.action_points_ids:
        raise Arcor2Exception("AP has invalid parent ID (not an object or another AP).")


@scene_needed
@project_needed
async def add_action_point_cb(req: srpc.p.AddActionPoint.Request, ui: WsClient) -> None:

    assert glob.SCENE
    assert glob.PROJECT

    hlp.is_valid_identifier(req.args.name)
    unique_name(req.args.name, glob.PROJECT.action_points_names)
    check_ap_parent(req.args.parent)

    if req.dry_run:
        return None

    ap = glob.PROJECT.upsert_action_point(common.ActionPoint.uid(), req.args.name, req.args.position, req.args.parent)

    evt = sevts.p.ActionPointChanged(ap)
    evt.change_type = Event.Type.ADD
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


@scene_needed
@project_needed
async def remove_action_point_cb(req: srpc.p.RemoveActionPoint.Request, ui: WsClient) -> None:

    assert glob.PROJECT

    ap = glob.PROJECT.bare_action_point(req.args.id)

    for proj_ap in glob.PROJECT.action_points_with_parent:
        if proj_ap.parent == ap.id:
            raise Arcor2Exception(f"Can't remove parent of '{proj_ap.name}' AP.")

    ap_action_ids = glob.PROJECT.ap_action_ids(ap.id)

    # check if AP's actions aren't involved in logic
    # TODO 'force' param to remove logical connections?
    for logic in glob.PROJECT.logic:
        if (
            logic.start in ap_action_ids
            or logic.end in ap_action_ids
            or (logic.condition and logic.condition.parse_what().action_id in ap_action_ids)
        ):
            raise Arcor2Exception("Remove logic connections first.")

    for act in glob.PROJECT.actions:

        if act.id in ap_action_ids:
            continue

        for param in act.parameters:
            if param.type == common.ActionParameter.TypeEnum.LINK:
                parsed_link = param.parse_link()
                linking_action = glob.PROJECT.action(parsed_link.action_id)
                if parsed_link.action_id in ap_action_ids:
                    raise Arcor2Exception(f"Result of '{act.name}' is linked from '{linking_action.name}'.")

            if not param.is_value():
                continue

            for joints in glob.PROJECT.ap_joints(ap.id):
                if plugin_from_type_name(param.type).uses_robot_joints(glob.PROJECT, act.id, param.name, joints.id):
                    raise Arcor2Exception(f"Joints {joints.name} used in action {act.name} (parameter {param.name}).")

            for ori in glob.PROJECT.ap_orientations(ap.id):
                if plugin_from_type_name(param.type).uses_orientation(glob.PROJECT, act.id, param.name, ori.id):
                    raise Arcor2Exception(f"Orientation {ori.name} used in action {act.name} (parameter {param.name}).")

            # TODO some hypothetical parameter type could use just bare ActionPoint (its position)

    if req.dry_run:
        return None

    glob.PROJECT.remove_action_point(req.args.id)

    evt = sevts.p.ActionPointChanged(ap)
    evt.change_type = Event.Type.REMOVE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


@scene_needed
@project_needed
async def copy_action_point_cb(req: srpc.p.CopyActionPoint.Request, ui: WsClient) -> None:

    assert glob.PROJECT

    def make_name_unique(orig_name: str, names: Set[str]) -> str:

        cnt = 1
        name = orig_name

        while name in names:
            name = f"{orig_name}_{cnt}"
            cnt += 1

        return name

    async def copy_action_point(
        orig_ap: common.BareActionPoint,
        new_parent_id: Optional[str] = None,
        position: Optional[common.Position] = None,
    ) -> None:

        assert glob.PROJECT

        ap = glob.PROJECT.upsert_action_point(
            common.ActionPoint.uid(),
            make_name_unique(f"{orig_ap.name}_copy", glob.PROJECT.action_points_names),
            orig_ap.position if position is None else position,
            orig_ap.parent if new_parent_id is None else new_parent_id,
        )

        ap_added_evt = sevts.p.ActionPointChanged(ap)
        ap_added_evt.change_type = Event.Type.ADD
        await notif.broadcast_event(ap_added_evt)

        for ori in glob.PROJECT.ap_orientations(orig_ap.id):
            new_ori = ori.copy()
            old_ori_to_new_ori[ori.id] = new_ori.id
            glob.PROJECT.upsert_orientation(ap.id, new_ori)

            ori_added_evt = sevts.p.OrientationChanged(new_ori)
            ori_added_evt.change_type = Event.Type.ADD
            ori_added_evt.parent_id = ap.id
            await notif.broadcast_event(ori_added_evt)

        for joints in glob.PROJECT.ap_joints(orig_ap.id):
            new_joints = joints.copy()
            glob.PROJECT.upsert_joints(ap.id, new_joints)

            joints_added_evt = sevts.p.JointsChanged(new_joints)
            joints_added_evt.change_type = Event.Type.ADD
            joints_added_evt.parent_id = ap.id
            await notif.broadcast_event(joints_added_evt)

        action_names = glob.PROJECT.action_names  # action name has to be globally unique
        for act in glob.PROJECT.ap_actions(orig_ap.id):
            new_act = act.copy()
            new_act.name = make_name_unique(f"{act.name}_copy", action_names)
            glob.PROJECT.upsert_action(ap.id, new_act)

            for param in new_act.parameters:

                if param.type != PosePlugin.type_name():
                    continue

                old_ori_id = PosePlugin.orientation_id(glob.PROJECT, new_act.id, param.name)

                # TODO this is hacky - plugins are missing methods to set/update parameters
                import json

                # TODO this won't work if action on AP is using orientation from AP's descendant
                #  ...which is not in the mapping yet
                try:
                    param.value = json.dumps(old_ori_to_new_ori[old_ori_id])
                except KeyError:
                    glob.logger.error(f"Failed to find a new orientation ID for {old_ori_id}.")

            action_added_evt = sevts.p.ActionChanged(new_act)
            action_added_evt.change_type = Event.Type.ADD
            action_added_evt.parent_id = ap.id
            await notif.broadcast_event(action_added_evt)

        for child_ap in glob.PROJECT.action_points_with_parent:
            if child_ap.parent == orig_ap.id:
                await copy_action_point(child_ap, ap.id)

    original_ap = glob.PROJECT.bare_action_point(req.args.id)

    if req.dry_run:
        return

    old_ori_to_new_ori: Dict[str, str] = {}
    asyncio.ensure_future(copy_action_point(original_ap, position=req.args.position))


@scene_needed
@project_needed
async def add_action_cb(req: srpc.p.AddAction.Request, ui: WsClient) -> None:

    assert glob.PROJECT
    assert glob.SCENE

    ap = glob.PROJECT.bare_action_point(req.args.action_point_id)

    unique_name(req.args.name, glob.PROJECT.action_names)

    new_action = common.Action(req.args.name, req.args.type, parameters=req.args.parameters, flows=req.args.flows)

    action_meta = find_object_action(glob.SCENE, new_action)

    updated_project = copy.deepcopy(glob.PROJECT)
    updated_project.upsert_action(req.args.action_point_id, new_action)

    check_flows(updated_project, new_action, action_meta)
    check_action_params(glob.SCENE, updated_project, new_action, action_meta)

    if req.dry_run:
        return None

    glob.PROJECT.upsert_action(ap.id, new_action)

    evt = sevts.p.ActionChanged(new_action)
    evt.change_type = Event.Type.ADD
    evt.parent_id = ap.id
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


@scene_needed
@project_needed
async def update_action_cb(req: srpc.p.UpdateAction.Request, ui: WsClient) -> None:

    assert glob.PROJECT
    assert glob.SCENE

    updated_project = copy.deepcopy(glob.PROJECT)

    updated_action = updated_project.action(req.args.action_id)

    if req.args.parameters is not None:
        updated_action.parameters = req.args.parameters
    if req.args.flows is not None:
        updated_action.flows = req.args.flows

    updated_action_meta = find_object_action(glob.SCENE, updated_action)

    check_flows(updated_project, updated_action, updated_action_meta)
    check_action_params(glob.SCENE, updated_project, updated_action, updated_action_meta)

    if req.dry_run:
        return None

    orig_action = glob.PROJECT.action(req.args.action_id)
    orig_action.parameters = updated_action.parameters
    glob.PROJECT.update_modified()

    evt = sevts.p.ActionChanged(updated_action)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


def check_action_usage(action: common.Action) -> None:

    assert glob.PROJECT

    # check parameters
    for act in glob.PROJECT.actions:
        for param in act.parameters:
            if param.type == common.ActionParameter.TypeEnum.LINK:
                link = param.parse_link()
                if action.id == link.action_id:
                    raise Arcor2Exception(f"Action output used as parameter of {act.name}/{param.name}.")

    # check logic
    for log in glob.PROJECT.logic:

        if log.start == action.id or log.end == action.id:
            raise Arcor2Exception("Action used in logic.")

        if log.condition:
            action_id, _, _ = log.condition.what.split("/")

            if action_id == action.id:
                raise Arcor2Exception("Action used in condition.")


@scene_needed
@project_needed
async def remove_action_cb(req: srpc.p.RemoveAction.Request, ui: WsClient) -> None:

    assert glob.PROJECT
    assert glob.SCENE

    ap, action = glob.PROJECT.action_point_and_action(req.args.id)
    check_action_usage(action)

    if req.dry_run:
        return None

    glob.PROJECT.remove_action(req.args.id)
    project.remove_prev_result(action.id)

    evt = sevts.p.ActionChanged(action.bare)
    evt.change_type = Event.Type.REMOVE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


def check_logic_item(parent: LogicContainer, logic_item: common.LogicItem) -> None:
    """Checks if newly added/updated ProjectLogicItem is ok.

    :param parent:
    :param logic_item:
    :return:
    """

    assert glob.SCENE

    action_ids = parent.action_ids()

    if logic_item.start == common.LogicItem.START and logic_item.end == common.LogicItem.END:
        raise Arcor2Exception("This does not make sense.")

    if logic_item.start != common.LogicItem.START:

        start_action_id, start_flow = logic_item.parse_start()

        if start_action_id == logic_item.end:
            raise Arcor2Exception("Start and end can't be the same.")

        if start_action_id not in action_ids:
            raise Arcor2Exception("Logic item has unknown start.")

        if start_flow != "default":
            raise Arcor2Exception("Only flow 'default' is supported so far.'")

    if logic_item.end != common.LogicItem.END:

        if logic_item.end not in action_ids:
            raise Arcor2Exception("Logic item has unknown end.")

    if logic_item.condition is not None:

        what = logic_item.condition.parse_what()
        action = parent.action(what.action_id)  # action that produced the result which we depend on here
        flow = action.flow(what.flow_name)
        try:
            flow.outputs[what.output_index]
        except IndexError:
            raise Arcor2Exception(f"Flow {flow.type} does not have output with index {what.output_index}.")

        action_meta = find_object_action(glob.SCENE, action)

        try:
            return_type = action_meta.returns[what.output_index]
        except IndexError:
            raise Arcor2Exception(f"Invalid output index {what.output_index} for action {action_meta.name}.")

        return_type_plugin = plugin_from_type_name(return_type)

        if not return_type_plugin.COUNTABLE:
            raise Arcor2Exception(f"Output of type {return_type} can't be branched.")

        # TODO for now there is only support for bool
        if return_type_plugin.type() != bool:
            raise Arcor2Exception("Unsupported condition type.")

        # check that condition value is ok, actual value is not interesting
        # TODO perform this check using plugin
        import json

        if not isinstance(json.loads(logic_item.condition.value), bool):
            raise Arcor2Exception("Invalid condition value.")

    for existing_item in parent.logic:

        if existing_item.id == logic_item.id:  # item is updated
            continue

        if logic_item.start == logic_item.START and existing_item.start == logic_item.START:
            raise Arcor2Exception("START already defined.")

        if logic_item.start == existing_item.start:

            if None in (logic_item.condition, existing_item.condition):
                raise Arcor2Exception("Two junctions has the same start action without condition.")

            # when there are more logical connections from A to B, their condition values must be different
            if logic_item.condition == existing_item.condition:
                raise Arcor2Exception("Two junctions with the same start should have different conditions.")

        if logic_item.end == existing_item.end:
            if logic_item.start == existing_item.start:
                raise Arcor2Exception("Junctions can't have the same start and end.")


@scene_needed
@project_needed
async def add_logic_item_cb(req: srpc.p.AddLogicItem.Request, ui: WsClient) -> None:

    assert glob.PROJECT
    assert glob.SCENE

    logic_item = common.LogicItem(req.args.start, req.args.end, req.args.condition)
    check_logic_item(glob.PROJECT, logic_item)

    if logic_item.start != logic_item.START:
        updated_project = copy.deepcopy(glob.PROJECT)
        updated_project.upsert_logic_item(logic_item)
        check_for_loops(updated_project, logic_item.parse_start().start_action_id)

    if req.dry_run:
        return

    glob.PROJECT.upsert_logic_item(logic_item)

    evt = sevts.p.LogicItemChanged(logic_item)
    evt.change_type = Event.Type.ADD
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


@scene_needed
@project_needed
async def update_logic_item_cb(req: srpc.p.UpdateLogicItem.Request, ui: WsClient) -> None:

    assert glob.PROJECT
    assert glob.SCENE

    updated_project = copy.deepcopy(glob.PROJECT)
    updated_logic_item = updated_project.logic_item(req.args.logic_item_id)

    updated_logic_item.start = req.args.start
    updated_logic_item.end = req.args.end
    updated_logic_item.condition = req.args.condition

    check_logic_item(updated_project, updated_logic_item)

    if updated_logic_item.start != updated_logic_item.START:
        check_for_loops(updated_project, updated_logic_item.parse_start().start_action_id)

    if req.dry_run:
        return

    glob.PROJECT.upsert_logic_item(updated_logic_item)

    evt = sevts.p.LogicItemChanged(updated_logic_item)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


@scene_needed
@project_needed
async def remove_logic_item_cb(req: srpc.p.RemoveLogicItem.Request, ui: WsClient) -> None:

    assert glob.PROJECT
    assert glob.SCENE

    logic_item = glob.PROJECT.logic_item(req.args.logic_item_id)

    # TODO is it necessary to check something here?
    glob.PROJECT.remove_logic_item(req.args.logic_item_id)

    evt = sevts.p.LogicItemChanged(logic_item)
    evt.change_type = Event.Type.REMOVE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


def check_constant(constant: common.ProjectConstant) -> None:

    assert glob.PROJECT

    hlp.is_valid_identifier(constant.name)

    for const in glob.PROJECT.constants:

        if constant.id == const.id:
            continue

        if constant.name == const.name:
            raise Arcor2Exception("Name has to be unique.")

    # TODO check using (constant?) plugin
    import json

    val = json.loads(constant.value)

    if not isinstance(val, (int, float, str, bool)):
        raise Arcor2Exception("Only basic types are supported so far.")


@scene_needed
@project_needed
async def add_constant_cb(req: srpc.p.AddConstant.Request, ui: WsClient) -> None:

    assert glob.PROJECT
    assert glob.SCENE

    const = common.ProjectConstant(req.args.name, req.args.type, req.args.value)
    check_constant(const)

    if req.dry_run:
        return

    glob.PROJECT.upsert_constant(const)

    evt = sevts.p.ProjectConstantChanged(const)
    evt.change_type = Event.Type.ADD
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


@scene_needed
@project_needed
async def update_constant_cb(req: srpc.p.UpdateConstant.Request, ui: WsClient) -> None:

    assert glob.PROJECT
    assert glob.SCENE

    const = glob.PROJECT.constant(req.args.constant_id)

    updated_constant = copy.deepcopy(const)

    if req.args.name is not None:
        updated_constant.name = req.args.name
    if req.args.value is not None:
        updated_constant.value = req.args.value

    check_constant(const)

    if req.dry_run:
        return

    glob.PROJECT.upsert_constant(updated_constant)

    evt = sevts.p.ProjectConstantChanged(const)
    evt.change_type = Event.Type.UPDATE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


@scene_needed
@project_needed
async def remove_constant_cb(req: srpc.p.RemoveConstant.Request, ui: WsClient) -> None:

    assert glob.PROJECT
    assert glob.SCENE

    const = glob.PROJECT.constant(req.args.constant_id)

    # check for usage
    for act in glob.PROJECT.actions:
        for param in act.parameters:
            if param.type == common.ActionParameter.TypeEnum.CONSTANT and param.str_from_value() == const.id:
                raise Arcor2Exception("Constant used as action parameter.")

    if req.dry_run:
        return

    glob.PROJECT.remove_constant(const.id)

    evt = sevts.p.ProjectConstantChanged(const)
    evt.change_type = Event.Type.REMOVE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


@no_project
async def delete_project_cb(req: srpc.p.DeleteProject.Request, ui: WsClient) -> None:

    project = UpdateableCachedProject(await storage.get_project(req.args.id))
    await storage.delete_project(req.args.id)

    evt = sevts.p.ProjectChanged(project.bare)
    evt.change_type = Event.Type.REMOVE
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def rename_project_cb(req: srpc.p.RenameProject.Request, ui: WsClient) -> None:

    unique_name(req.args.new_name, (await project_names()))

    if req.dry_run:
        return None

    async with managed_project(req.args.project_id) as project:

        project.name = req.args.new_name
        project.update_modified()

        evt = sevts.p.ProjectChanged(project.bare)
        evt.change_type = Event.Type.UPDATE_BASE
        asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def copy_project_cb(req: srpc.p.CopyProject.Request, ui: WsClient) -> None:

    unique_name(req.args.target_name, (await project_names()))

    if req.dry_run:
        return None

    async with managed_project(req.args.source_id, make_copy=True) as project:

        project.name = req.args.target_name

        evt = sevts.p.ProjectChanged(project.bare)
        evt.change_type = Event.Type.UPDATE_BASE
        asyncio.ensure_future(notif.broadcast_event(evt))

    return None


async def update_project_description_cb(req: srpc.p.UpdateProjectDescription.Request, ui: WsClient) -> None:

    async with managed_project(req.args.project_id) as project:

        project.desc = req.args.new_description
        project.update_modified()

        evt = sevts.p.ProjectChanged(project.bare)
        evt.change_type = Event.Type.UPDATE_BASE
        asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def update_project_has_logic_cb(req: srpc.p.UpdateProjectHasLogic.Request, ui: WsClient) -> None:

    async with managed_project(req.args.project_id) as project:

        if project.has_logic and not req.args.new_has_logic:
            project.clear_logic()

            """
            TODO
            if glob.PROJECT and glob.PROJECT.id == req.args.project_id:
            ...send remove event for each logic item?
            """

        project.has_logic = req.args.new_has_logic
        project.update_modified()

        evt = sevts.p.ProjectChanged(project.bare)
        evt.change_type = Event.Type.UPDATE_BASE
        asyncio.ensure_future(notif.broadcast_event(evt))
    return None


@project_needed
async def rename_action_point_joints_cb(req: srpc.p.RenameActionPointJoints.Request, ui: WsClient) -> None:

    assert glob.PROJECT

    ap, joints = glob.PROJECT.ap_and_joints(req.args.joints_id)
    hlp.is_valid_identifier(req.args.new_name)
    unique_name(req.args.new_name, glob.PROJECT.ap_joint_names(ap.id))

    if req.dry_run:
        return None

    joints.name = req.args.new_name
    glob.PROJECT.update_modified()

    evt = sevts.p.JointsChanged(joints)
    evt.change_type = Event.Type.UPDATE_BASE
    asyncio.ensure_future(notif.broadcast_event(evt))

    return None


@project_needed
async def rename_action_point_orientation_cb(req: srpc.p.RenameActionPointOrientation.Request, ui: WsClient) -> None:

    assert glob.PROJECT

    ap, ori = glob.PROJECT.bare_ap_and_orientation(req.args.orientation_id)
    hlp.is_valid_identifier(req.args.new_name)
    unique_name(req.args.new_name, glob.PROJECT.ap_orientation_names(ap.id))

    if req.dry_run:
        return None

    ori.name = req.args.new_name
    glob.PROJECT.update_modified()

    evt = sevts.p.OrientationChanged(ori)
    evt.change_type = Event.Type.UPDATE_BASE
    asyncio.ensure_future(notif.broadcast_event(evt))

    return None


@project_needed
async def rename_action_cb(req: srpc.p.RenameAction.Request, ui: WsClient) -> None:

    assert glob.PROJECT

    unique_name(req.args.new_name, glob.PROJECT.action_names)

    if req.dry_run:
        return None

    act = glob.PROJECT.action(req.args.action_id)
    act.name = req.args.new_name

    glob.PROJECT.update_modified()

    evt = sevts.p.ActionChanged(act)
    evt.change_type = Event.Type.UPDATE_BASE
    asyncio.ensure_future(notif.broadcast_event(evt))

    return None



import functools
from typing import Any, Callable, Coroutine, TypeVar, cast

from arcor2.exceptions import Arcor2Exception
from arcor2_arserver import globals as glob

F = TypeVar("F", bound=Callable[..., Coroutine[Any, Any, Any]])


def no_scene(coro: F) -> F:
    @functools.wraps(coro)
    async def async_wrapper(*args, **kwargs) -> Any:

        if glob.SCENE:
            raise Arcor2Exception("Scene has to be closed first.")
        return await coro(*args, **kwargs)

    return cast(F, async_wrapper)


def scene_needed(coro: F) -> F:
    @functools.wraps(coro)
    async def async_wrapper(*args, **kwargs) -> Any:

        if glob.SCENE is None or not glob.SCENE.id:
            raise Arcor2Exception("Scene not opened or has invalid id.")
        return await coro(*args, **kwargs)

    return cast(F, async_wrapper)


def no_project(coro: F) -> F:
    @functools.wraps(coro)
    async def async_wrapper(*args, **kwargs) -> Any:
        if glob.PROJECT:
            raise Arcor2Exception("Not available during project editing.")
        return await coro(*args, **kwargs)

    return cast(F, async_wrapper)


def project_needed(coro: F) -> F:
    @functools.wraps(coro)
    async def async_wrapper(*args, **kwargs) -> Any:

        if glob.PROJECT is None or not glob.PROJECT.id:
            raise Arcor2Exception("Project not opened or has invalid id.")
        return await coro(*args, **kwargs)

    return cast(F, async_wrapper)



from ast import AST, Assign, Call, ClassDef, ImportFrom, Module, Name, NameConstant, Pass, Store, alias

import humps

import arcor2.object_types
from arcor2.exceptions import Arcor2NotImplemented
from arcor2.object_types.utils import built_in_types_names
from arcor2.source import SourceException
from arcor2.source.utils import find_function, find_raises, get_name
from arcor2_arserver_data.objects import ObjectTypeMeta


# TODO could this be done like this https://stackoverflow.com/a/9269964/3142796 ??
def new_object_type(parent: ObjectTypeMeta, child: ObjectTypeMeta) -> AST:

    assert parent.type == child.base

    tree = Module(body=[], type_ignores=[])

    if parent.type in built_in_types_names():
        import_from = arcor2.object_types.abstract.__name__
    else:
        import_from = f".{humps.depascalize(parent.type)}"

    tree.body.append(ImportFrom(module=import_from, names=[alias(name=parent.type, asname=None)], level=0))

    c = ClassDef(
        name=child.type,
        bases=[get_name(parent.type)],
        keywords=[],
        body=[
            Assign(
                targets=[Name(id="_ABSTRACT", ctx=Store())],
                value=NameConstant(value=False, kind=None),
                type_comment=None,
            )
        ],
        decorator_list=[],
    )

    # TODO add docstring with description (if provided)
    c.body.append(Pass())

    tree.body.append(c)

    return tree


def function_implemented(tree: AST, func_name: str) -> bool:
    """Body of unimplemented function (e.g. object/service feature) contains
    only 'raise Arcor2NotImplemented()'.

    :param tree:
    :return:
    """

    try:
        func_def = find_function(func_name, tree)
    except SourceException:
        return False

    raises = find_raises(func_def)

    if len(raises) != 1:
        return True

    exc = raises[0].exc

    if isinstance(exc, Call):
        # raise Arcor2NotImplemented("something")

        assert isinstance(exc.func, Name)

        if exc.func.id == Arcor2NotImplemented.__name__:
            return False

    if isinstance(exc, Name) and exc.id == Arcor2NotImplemented.__name__:
        # raise Arcor2NotImplemented
        return False

    return True



from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import Camera
from arcor2_arserver.scene import get_instance


def get_camera_instance(obj_id: str) -> Camera:

    obj = get_instance(obj_id)

    if not isinstance(obj, Camera):
        raise Arcor2Exception("Not a camera.")

    return obj



import asyncio
import math
import time
from typing import Awaitable, Callable, Dict, List, Optional

import numpy as np
from arcor2_calibration_data import client as calib_client
from arcor2_calibration_data.client import CalibrateRobotArgs
from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import transformations as tr
from arcor2 import ws_server
from arcor2.clients.persistent_storage import URL as ps_url
from arcor2.data import common
from arcor2.exceptions import Arcor2Exception
from arcor2.helpers import run_in_executor
from arcor2.object_types.abstract import Camera, Robot
from arcor2_arserver import camera
from arcor2_arserver import globals as glob
from arcor2_arserver import notifications as notif
from arcor2_arserver import objects_actions as osa
from arcor2_arserver import robot
from arcor2_arserver.decorators import project_needed, scene_needed
from arcor2_arserver.scene import ensure_scene_started, scene_started, update_scene_object_pose
from arcor2_arserver_data import events as sevts
from arcor2_arserver_data import rpc as srpc
from arcor2_arserver_data.events.common import ProcessState
from arcor2_arserver_data.events.robot import HandTeachingMode

RBT_CALIB = "RobotCalibration"

TaskDict = Dict[str, asyncio.Task]

ROBOT_JOINTS_TASKS: TaskDict = {}
EEF_POSE_TASKS: TaskDict = {}

EVENT_PERIOD = 0.1


async def robot_joints_event(robot_id: str) -> None:

    global ROBOT_JOINTS_TASKS

    glob.logger.info(f"Sending '{sevts.r.RobotJoints.__name__}' for robot '{robot_id}' started.")
    while scene_started() and glob.ROBOT_JOINTS_REGISTERED_UIS[robot_id]:

        start = time.monotonic()

        try:
            evt = sevts.r.RobotJoints(sevts.r.RobotJoints.Data(robot_id, (await robot.get_robot_joints(robot_id))))
        except Arcor2Exception as e:
            glob.logger.error(f"Failed to get joints for {robot_id}. {str(e)}")
            break

        evt_json = evt.to_json()
        await asyncio.gather(
            *[ws_server.send_json_to_client(ui, evt_json) for ui in glob.ROBOT_JOINTS_REGISTERED_UIS[robot_id]]
        )

        end = time.monotonic()
        await asyncio.sleep(EVENT_PERIOD - (end - start))

    del ROBOT_JOINTS_TASKS[robot_id]

    # TODO notify UIs that registration was cancelled
    del glob.ROBOT_JOINTS_REGISTERED_UIS[robot_id]

    glob.logger.info(f"Sending '{sevts.r.RobotJoints.__name__}' for robot '{robot_id}' stopped.")


async def eef_pose(robot_id: str, eef_id: str) -> sevts.r.RobotEef.Data.EefPose:

    return sevts.r.RobotEef.Data.EefPose(eef_id, (await robot.get_end_effector_pose(robot_id, eef_id)))


async def robot_eef_pose_event(robot_id: str) -> None:

    global EEF_POSE_TASKS

    glob.logger.info(f"Sending '{sevts.r.RobotEef.__name__}' for robot '{robot_id}' started.")

    while scene_started() and glob.ROBOT_EEF_REGISTERED_UIS[robot_id]:

        start = time.monotonic()

        evt = sevts.r.RobotEef(sevts.r.RobotEef.Data(robot_id))

        try:
            evt.data.end_effectors = await asyncio.gather(
                *[eef_pose(robot_id, eef_id) for eef_id in (await robot.get_end_effectors(robot_id))]
            )
        except Arcor2Exception as e:
            glob.logger.error(f"Failed to get eef pose for {robot_id}. {str(e)}")
            break

        evt_json = evt.to_json()
        await asyncio.gather(
            *[ws_server.send_json_to_client(ui, evt_json) for ui in glob.ROBOT_EEF_REGISTERED_UIS[robot_id]]
        )

        end = time.monotonic()
        await asyncio.sleep(EVENT_PERIOD - (end - start))

    del EEF_POSE_TASKS[robot_id]

    # TODO notify UIs that registration was cancelled
    del glob.ROBOT_EEF_REGISTERED_UIS[robot_id]

    glob.logger.info(f"Sending '{sevts.r.RobotEef.__name__}' for robot '{robot_id}' stopped.")


async def get_robot_meta_cb(req: srpc.r.GetRobotMeta.Request, ui: WsClient) -> srpc.r.GetRobotMeta.Response:

    return srpc.r.GetRobotMeta.Response(
        data=[obj.robot_meta for obj in glob.OBJECT_TYPES.values() if obj.robot_meta is not None]
    )


@scene_needed
async def get_robot_joints_cb(req: srpc.r.GetRobotJoints.Request, ui: WsClient) -> srpc.r.GetRobotJoints.Response:

    ensure_scene_started()
    return srpc.r.GetRobotJoints.Response(data=await robot.get_robot_joints(req.args.robot_id))


@scene_needed
async def get_end_effector_pose_cb(
    req: srpc.r.GetEndEffectorPose.Request, ui: WsClient
) -> srpc.r.GetEndEffectorPose.Response:

    ensure_scene_started()
    return srpc.r.GetEndEffectorPose.Response(
        data=await robot.get_end_effector_pose(req.args.robot_id, req.args.end_effector_id)
    )


@scene_needed
async def get_end_effectors_cb(req: srpc.r.GetEndEffectors.Request, ui: WsClient) -> srpc.r.GetEndEffectors.Response:

    ensure_scene_started()
    return srpc.r.GetEndEffectors.Response(data=await robot.get_end_effectors(req.args.robot_id))


@scene_needed
async def get_grippers_cb(req: srpc.r.GetGrippers.Request, ui: WsClient) -> srpc.r.GetGrippers.Response:

    ensure_scene_started()
    return srpc.r.GetGrippers.Response(data=await robot.get_grippers(req.args.robot_id))


@scene_needed
async def get_suctions_cb(req: srpc.r.GetSuctions.Request, ui: WsClient) -> srpc.r.GetSuctions.Response:

    ensure_scene_started()
    return srpc.r.GetSuctions.Response(data=await robot.get_suctions(req.args.robot_id))


async def register(
    req: srpc.r.RegisterForRobotEvent.Request,
    ui: WsClient,
    tasks: TaskDict,
    reg_uis: glob.RegisteredUiDict,
    coro: Callable[[str], Awaitable[None]],
) -> None:

    if req.args.send:

        reg_uis[req.args.robot_id].add(ui)

        if req.args.robot_id not in tasks:
            # start task
            tasks[req.args.robot_id] = asyncio.create_task(coro(req.args.robot_id))

    else:
        try:
            reg_uis[req.args.robot_id].remove(ui)
        except KeyError as e:
            raise Arcor2Exception("Failed to unregister.") from e

        # cancel task if not needed anymore
        if not reg_uis[req.args.robot_id]:
            task = tasks[req.args.robot_id]

            if not task.cancelled():
                task.cancel()

            del tasks[req.args.robot_id]


@scene_needed
async def register_for_robot_event_cb(req: srpc.r.RegisterForRobotEvent.Request, ui: WsClient) -> None:

    ensure_scene_started()

    # check if robot exists
    await osa.get_robot_instance(req.args.robot_id)

    if req.args.what == req.args.RegisterEnum.JOINTS:
        await register(req, ui, ROBOT_JOINTS_TASKS, glob.ROBOT_JOINTS_REGISTERED_UIS, robot_joints_event)
    elif req.args.what == req.args.RegisterEnum.EEF_POSE:

        if not (await robot.get_end_effectors(req.args.robot_id)):
            raise Arcor2Exception("Robot does not have any end effector.")

        await register(req, ui, EEF_POSE_TASKS, glob.ROBOT_EEF_REGISTERED_UIS, robot_eef_pose_event)
    else:
        raise Arcor2Exception(f"Option '{req.args.what.value}' not implemented.")

    return None


async def check_feature(robot_id: str, feature_name: str) -> None:

    obj_type = glob.OBJECT_TYPES[(await osa.get_robot_instance(robot_id)).__class__.__name__]

    if obj_type.robot_meta is None:
        raise Arcor2Exception("Not a robot.")

    if not getattr(obj_type.robot_meta.features, feature_name):
        raise Arcor2Exception(f"Robot does not support '{feature_name}' feature.")


@scene_needed
async def move_to_pose_cb(req: srpc.r.MoveToPose.Request, ui: WsClient) -> None:

    ensure_scene_started()
    await check_feature(req.args.robot_id, Robot.move_to_pose.__name__)
    await robot.check_robot_before_move(req.args.robot_id)

    if (req.args.position is None) != (req.args.orientation is None):

        target_pose = await robot.get_end_effector_pose(req.args.robot_id, req.args.end_effector_id)

        if req.args.position:
            target_pose.position = req.args.position
        elif req.args.orientation:
            target_pose.orientation = req.args.orientation

    elif req.args.position is not None and req.args.orientation is not None:
        target_pose = common.Pose(req.args.position, req.args.orientation)
    else:
        raise Arcor2Exception("Position or orientation should be given.")

    # TODO check if the target pose is reachable (dry_run)
    asyncio.ensure_future(
        robot.move_to_pose(req.args.robot_id, req.args.end_effector_id, target_pose, req.args.speed, req.args.safe)
    )


@scene_needed
async def move_to_joints_cb(req: srpc.r.MoveToJoints.Request, ui: WsClient) -> None:

    ensure_scene_started()
    await check_feature(req.args.robot_id, Robot.move_to_joints.__name__)
    await robot.check_robot_before_move(req.args.robot_id)
    asyncio.ensure_future(robot.move_to_joints(req.args.robot_id, req.args.joints, req.args.speed, req.args.safe))


@scene_needed
async def stop_robot_cb(req: srpc.r.StopRobot.Request, ui: WsClient) -> None:

    ensure_scene_started()
    await check_feature(req.args.robot_id, Robot.stop.__name__)
    await robot.stop(req.args.robot_id)


@scene_needed
@project_needed
async def move_to_action_point_cb(req: srpc.r.MoveToActionPoint.Request, ui: WsClient) -> None:

    ensure_scene_started()
    await robot.check_robot_before_move(req.args.robot_id)

    assert glob.SCENE
    assert glob.PROJECT

    if (req.args.orientation_id is None) == (req.args.joints_id is None):
        raise Arcor2Exception("Set orientation or joints. Not both.")

    if req.args.orientation_id:

        await check_feature(req.args.robot_id, Robot.move_to_pose.__name__)

        if req.args.end_effector_id is None:
            raise Arcor2Exception("eef id has to be set.")

        pose = tr.abs_pose_from_ap_orientation(glob.SCENE, glob.PROJECT, req.args.orientation_id)

        # TODO check if the target pose is reachable (dry_run)
        asyncio.ensure_future(
            robot.move_to_ap_orientation(
                req.args.robot_id,
                req.args.end_effector_id,
                pose,
                req.args.speed,
                req.args.orientation_id,
                req.args.safe,
            )
        )

    elif req.args.joints_id:

        await check_feature(req.args.robot_id, Robot.move_to_joints.__name__)

        joints = glob.PROJECT.joints(req.args.joints_id)

        # TODO check if the joints are within limits and reachable (dry_run)
        asyncio.ensure_future(
            robot.move_to_ap_joints(req.args.robot_id, joints.joints, req.args.speed, req.args.joints_id, req.args.safe)
        )


@scene_needed
async def ik_cb(req: srpc.r.InverseKinematics.Request, ui: WsClient) -> srpc.r.InverseKinematics.Response:

    ensure_scene_started()
    await check_feature(req.args.robot_id, Robot.inverse_kinematics.__name__)

    joints = await robot.ik(
        req.args.robot_id, req.args.end_effector_id, req.args.pose, req.args.start_joints, req.args.avoid_collisions
    )
    resp = srpc.r.InverseKinematics.Response()
    resp.data = joints
    return resp


@scene_needed
async def fk_cb(req: srpc.r.ForwardKinematics.Request, ui: WsClient) -> srpc.r.ForwardKinematics.Response:

    ensure_scene_started()
    await check_feature(req.args.robot_id, Robot.forward_kinematics.__name__)

    pose = await robot.fk(req.args.robot_id, req.args.end_effector_id, req.args.joints)
    resp = srpc.r.ForwardKinematics.Response()
    resp.data = pose
    return resp


async def calibrate_robot(robot_inst: Robot, camera_inst: Camera, move_to_calibration_pose: bool) -> None:

    assert glob.SCENE
    assert camera_inst.color_camera_params

    # TODO it should not be possible to close the scene during this process

    await notif.broadcast_event(ProcessState(ProcessState.Data(RBT_CALIB, ProcessState.Data.StateEnum.Started)))

    try:

        if move_to_calibration_pose:
            await run_in_executor(robot_inst.move_to_calibration_pose)
        robot_joints = await run_in_executor(robot_inst.robot_joints)
        depth_image = await run_in_executor(camera_inst.depth_image, 128)

        args = CalibrateRobotArgs(
            robot_joints,
            robot_inst.pose,
            camera_inst.pose,
            camera_inst.color_camera_params,
            f"{ps_url}/models/{robot_inst.urdf_package_name}/mesh/file",
        )

        new_pose = await run_in_executor(calib_client.calibrate_robot, args, depth_image)

    except Arcor2Exception as e:
        await notif.broadcast_event(
            ProcessState(ProcessState.Data(RBT_CALIB, ProcessState.Data.StateEnum.Failed, str(e)))
        )
        glob.logger.exception("Failed to calibrate the robot.")
        return

    await update_scene_object_pose(glob.SCENE.object(robot_inst.id), new_pose, robot_inst)
    await notif.broadcast_event(ProcessState(ProcessState.Data(RBT_CALIB, ProcessState.Data.StateEnum.Finished)))


@scene_needed
async def calibrate_robot_cb(req: srpc.r.CalibrateRobot.Request, ui: WsClient) -> None:

    ensure_scene_started()
    robot_inst = await osa.get_robot_instance(req.args.robot_id)

    if not robot_inst.urdf_package_name:
        raise Arcor2Exception("Robot with model required!")

    if req.args.camera_id:
        camera_inst = camera.get_camera_instance(req.args.camera_id)
    else:
        for obj in glob.SCENE_OBJECT_INSTANCES.values():
            if isinstance(obj, Camera):
                camera_inst = obj
                break
        else:
            raise Arcor2Exception("No camera found.")

    if camera_inst.color_camera_params is None:
        raise Arcor2Exception("Calibrated camera required!")

    # TODO check camera features / check that it supports depth
    asyncio.ensure_future(calibrate_robot(robot_inst, camera_inst, req.args.move_to_calibration_pose))

    return None


@scene_needed
async def hand_teaching_mode_cb(req: srpc.r.HandTeachingMode.Request, ui: WsClient) -> None:

    ensure_scene_started()
    robot_inst = await osa.get_robot_instance(req.args.robot_id)

    otd = osa.get_obj_type_data(req.args.robot_id)
    assert otd.robot_meta is not None
    if not otd.robot_meta.features.hand_teaching:
        raise Arcor2Exception("Robot does not support hand teaching.")

    hand_teaching_mode = await run_in_executor(robot_inst.get_hand_teaching_mode)

    if req.args.enable == hand_teaching_mode:
        raise Arcor2Exception("That's the current state.")

    if req.dry_run:
        return

    await run_in_executor(robot_inst.set_hand_teaching_mode, req.args.enable)
    evt = HandTeachingMode(HandTeachingMode.Data(req.args.robot_id, req.args.enable))
    asyncio.ensure_future(notif.broadcast_event(evt))


@scene_needed
async def step_robot_eef_cb(req: srpc.r.StepRobotEef.Request, ui: WsClient) -> None:

    ensure_scene_started()
    await check_feature(req.args.robot_id, Robot.move_to_pose.__name__)
    await robot.check_robot_before_move(req.args.robot_id)

    robot_inst = await osa.get_robot_instance(req.args.robot_id)

    tp = await robot.get_end_effector_pose(req.args.robot_id, req.args.end_effector_id)

    if req.args.mode == req.args.mode.ROBOT:
        tp = tr.make_pose_rel(robot_inst.pose, tp)
    elif req.args.mode == req.args.mode.RELATIVE:
        assert req.args.pose
        tp = tr.make_pose_rel(req.args.pose, tp)
    elif req.args.mode == req.args.mode.USER:
        assert req.args.pose
        raise Arcor2Exception("Not supported yet.")

    if req.args.what == req.args.what.POSITION:
        if req.args.axis == req.args.axis.X:
            tp.position.x += req.args.step
        elif req.args.axis == req.args.axis.Y:
            tp.position.y += req.args.step
        elif req.args.axis == req.args.axis.Z:
            tp.position.z += req.args.step
    elif req.args.what == req.args.what.ORIENTATION:
        if req.args.axis == req.args.axis.X:
            tp.orientation *= common.Orientation.from_rotation_vector(x=req.args.step)
        elif req.args.axis == req.args.axis.Y:
            tp.orientation *= common.Orientation.from_rotation_vector(y=req.args.step)
        elif req.args.axis == req.args.axis.Z:
            tp.orientation *= common.Orientation.from_rotation_vector(z=req.args.step)

    if req.args.mode == req.args.mode.ROBOT:
        tp = tr.make_pose_abs(robot_inst.pose, tp)
    elif req.args.mode == req.args.mode.RELATIVE:
        assert req.args.pose
        tp = tr.make_pose_abs(req.args.pose, tp)

    await robot.check_reachability(req.args.robot_id, req.args.end_effector_id, tp, req.args.safe)

    if req.dry_run:
        return

    asyncio.ensure_future(
        robot.move_to_pose(req.args.robot_id, req.args.end_effector_id, tp, req.args.speed, req.args.safe)
    )


@scene_needed
async def set_eef_perpendicular_to_world_cb(req: srpc.r.SetEefPerpendicularToWorld.Request, ui: WsClient) -> None:

    ensure_scene_started()

    await check_feature(req.args.robot_id, Robot.move_to_pose.__name__)
    await check_feature(req.args.robot_id, Robot.inverse_kinematics.__name__)
    await robot.check_robot_before_move(req.args.robot_id)

    if req.dry_run:  # attempt to find suitable joints can take some time so it is not done for dry_run
        return

    tp = await robot.get_end_effector_pose(req.args.robot_id, req.args.end_effector_id)

    current_joints = await robot.get_robot_joints(req.args.robot_id)
    target_joints: Optional[List[common.Joint]] = None
    target_joints_diff: float = 0.0

    # TODO call ik in parallel?
    # select best (closest joint configuration) reachable pose
    for z_rot in np.linspace(-math.pi, math.pi, 360):
        tp.orientation = common.Orientation.from_rotation_vector(y=math.pi) * common.Orientation.from_rotation_vector(
            z=z_rot
        )
        try:
            joints = await robot.ik(req.args.robot_id, req.args.end_effector_id, tp, current_joints, req.args.safe)
        except Arcor2Exception:
            continue

        if not target_joints:
            target_joints = joints
            for f, b in zip(current_joints, target_joints):
                assert f.name == b.name
                target_joints_diff += (f.value - b.value) ** 2
        else:
            diff = 0.0
            for f, b in zip(current_joints, joints):
                assert f.name == b.name
                diff += (f.value - b.value) ** 2

            if diff < target_joints_diff:
                target_joints = joints
                target_joints_diff = diff

    if not target_joints:
        raise Arcor2Exception("Could not find reachable pose.")

    asyncio.ensure_future(robot.move_to_joints(req.args.robot_id, target_joints, req.args.speed, req.args.safe))



import asyncio
from typing import AsyncIterator, Dict, List, Optional, Set

from arcor2 import helpers as hlp
from arcor2.cached import CachedScene, UpdateableCachedScene
from arcor2.clients import aio_scene_service as scene_srv
from arcor2.data.common import Parameter, Pose, SceneObject
from arcor2.data.events import Event
from arcor2.data.object_type import Models
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import Generic, GenericWithPose, Robot
from arcor2.object_types.utils import settings_from_params
from arcor2_arserver import globals as glob
from arcor2_arserver import notifications as notif
from arcor2_arserver.clients import persistent_storage as storage
from arcor2_arserver.object_types.data import ObjectTypeData
from arcor2_arserver.objects_actions import get_object_types
from arcor2_arserver_data.events.common import ShowMainScreen
from arcor2_arserver_data.events.scene import SceneClosed, SceneObjectChanged, SceneState

# TODO maybe this could be property of ARServerScene(CachedScene)?
_scene_state: SceneState.Data.StateEnum = SceneState.Data.StateEnum.Stopped


async def update_scene_object_pose(
    obj: SceneObject, pose: Optional[Pose] = None, obj_inst: Optional[GenericWithPose] = None
) -> None:
    """Performs all necessary actions when pose of an object is updated.

    :param obj:
    :param pose:
    :param obj_inst:
    :return:
    """

    assert glob.SCENE

    if pose:
        # SceneObject pose was not updated before
        obj.pose = pose
    else:
        # SceneObject pose was already updated
        pose = obj.pose

    glob.SCENE.update_modified()

    evt = SceneObjectChanged(obj)
    evt.change_type = Event.Type.UPDATE
    await notif.broadcast_event(evt)

    glob.OBJECTS_WITH_UPDATED_POSE.add(obj.id)

    if scene_started():

        if obj_inst is None:
            inst = get_instance(obj.id)
            assert isinstance(inst, GenericWithPose)
            obj_inst = inst

        assert pose is not None

        # Object pose is property that might call scene service - that's why it has to be called using executor.
        await hlp.run_in_executor(setattr, obj_inst, "pose", pose)


async def set_scene_state(state: SceneState.Data.StateEnum, message: Optional[str] = None) -> None:

    global _scene_state
    _scene_state = state
    await notif.broadcast_event(SceneState(SceneState.Data(state, message)))


def get_scene_state() -> SceneState.Data.StateEnum:
    return _scene_state


def scene_started() -> bool:
    return _scene_state == SceneState.Data.StateEnum.Started


def can_modify_scene() -> None:
    """Raises exception if modifications to scene/project are not possible."""

    if _scene_state != SceneState.Data.StateEnum.Stopped:
        raise Arcor2Exception("Modifications can be only done offline.")


def ensure_scene_started() -> None:
    """" Raises exception if scene is not started."""

    if _scene_state != SceneState.Data.StateEnum.Started:
        raise Arcor2Exception("Scene offline.")


async def scenes() -> AsyncIterator[CachedScene]:

    for scene_id in (await storage.get_scenes()).items:
        yield CachedScene(await storage.get_scene(scene_id.id))


async def scene_names() -> Set[str]:
    return {scene.name for scene in (await storage.get_scenes()).items}


async def notify_scene_closed(scene_id: str) -> None:

    await notif.broadcast_event(SceneClosed())
    glob.MAIN_SCREEN = ShowMainScreen.Data(ShowMainScreen.Data.WhatEnum.ScenesList)
    await notif.broadcast_event(
        ShowMainScreen(data=ShowMainScreen.Data(ShowMainScreen.Data.WhatEnum.ScenesList, scene_id))
    )


def check_object_parameters(obj_type: ObjectTypeData, parameters: List[Parameter]) -> None:

    if {s.name for s in obj_type.meta.settings if s.default_value is None} > {s.name for s in parameters}:
        raise Arcor2Exception("Some required parameter is missing.")

    # TODO check types of parameters, ranges, etc.


def check_object(obj: SceneObject, new_one: bool = False) -> None:
    """Checks if object can be added into the scene."""

    assert glob.SCENE
    assert not obj.children

    if obj.type not in glob.OBJECT_TYPES:
        raise Arcor2Exception("Unknown object type.")

    obj_type = glob.OBJECT_TYPES[obj.type]

    if obj_type.meta.disabled:
        raise Arcor2Exception("Object type disabled.")

    check_object_parameters(obj_type, obj.parameters)

    # TODO check whether object needs parent and if so, if the parent is in scene and parent_id is set
    if obj_type.meta.needs_parent_type:
        pass

    if obj_type.meta.has_pose and obj.pose is None:
        raise Arcor2Exception("Object requires pose.")

    if not obj_type.meta.has_pose and obj.pose is not None:
        raise Arcor2Exception("Object do not have pose.")

    if obj_type.meta.abstract:
        raise Arcor2Exception("Cannot instantiate abstract type.")

    if new_one:

        if obj.id in glob.SCENE.object_ids:
            raise Arcor2Exception("Object/service with that id already exists.")

        if obj.name in glob.SCENE.object_names():
            raise Arcor2Exception("Name is already used.")

    hlp.is_valid_identifier(obj.name)


async def add_object_to_scene(obj: SceneObject, dry_run: bool = False) -> None:
    """

    :param obj:
    :param add_to_scene: Set to false to only create object instance and add its collision model (if any).
    :return:
    """

    assert glob.SCENE

    check_object(obj, new_one=True)

    if dry_run:
        return None

    glob.SCENE.upsert_object(obj)
    glob.logger.debug(f"Object {obj.id} ({obj.type}) added to scene.")


async def create_object_instance(obj: SceneObject, overrides: Optional[List[Parameter]] = None) -> None:

    obj_type = glob.OBJECT_TYPES[obj.type]

    # settings -> dataclass
    assert obj_type.type_def
    glob.logger.debug(
        f"Creating instance of {obj_type.type_def.__name__} with name {obj.name}. "
        f"Parameters: {obj.parameters}, overrides: {overrides}."
    )
    settings = settings_from_params(obj_type.type_def, obj.parameters, overrides)

    assert obj_type.type_def is not None

    try:

        try:

            if issubclass(obj_type.type_def, Robot):
                assert obj.pose is not None
                glob.SCENE_OBJECT_INSTANCES[obj.id] = await hlp.run_in_executor(
                    obj_type.type_def, obj.id, obj.name, obj.pose, settings
                )
            elif issubclass(obj_type.type_def, GenericWithPose):
                assert obj.pose is not None
                coll_model: Optional[Models] = None
                if obj_type.meta.object_model:
                    coll_model = obj_type.meta.object_model.model()

                glob.SCENE_OBJECT_INSTANCES[obj.id] = await hlp.run_in_executor(
                    obj_type.type_def, obj.id, obj.name, obj.pose, coll_model, settings
                )

            elif issubclass(obj_type.type_def, Generic):
                assert obj.pose is None
                glob.SCENE_OBJECT_INSTANCES[obj.id] = await hlp.run_in_executor(
                    obj_type.type_def, obj.id, obj.name, settings
                )

            else:
                raise Arcor2Exception("Object type with unknown base.")

        except (TypeError, ValueError) as e:  # catch some most often exceptions
            raise Arcor2Exception("Unhandled error.") from e

    except Arcor2Exception as e:
        # make the exception a bit more user-friendly by including the object's name
        raise Arcor2Exception(f"Failed to initialize {obj.name}. {str(e)}") from e

    return None


async def open_scene(scene_id: str) -> None:

    await get_object_types()
    asyncio.ensure_future(scene_srv.delete_all_collisions())
    glob.SCENE = UpdateableCachedScene(await storage.get_scene(scene_id))

    try:
        for obj in glob.SCENE.objects:
            check_object(obj)
    except Arcor2Exception as e:
        glob.SCENE = None
        raise Arcor2Exception(f"Failed to open scene. {str(e)}") from e


# TODO optional parameter for expected return type
def get_instance(obj_id: str) -> Generic:

    if obj_id not in glob.SCENE_OBJECT_INSTANCES:
        raise Arcor2Exception("Unknown object/service ID.")

    return glob.SCENE_OBJECT_INSTANCES[obj_id]


async def cleanup_object(obj: Generic) -> None:

    try:
        await hlp.run_in_executor(obj.cleanup)
    except Arcor2Exception as e:
        # make the exception a bit more user-friendly by including the object's name
        raise Arcor2Exception(f"Failed to cleanup {obj.name}. {str(e)}") from e


async def stop_scene(message: Optional[str] = None) -> None:
    """Destroys scene object instances."""

    glob.logger.info("Stopping the scene.")

    await set_scene_state(SceneState.Data.StateEnum.Stopping, message)

    if await scene_srv.started():
        try:
            await scene_srv.stop()
        except Arcor2Exception as e:
            glob.logger.exception("Failed to go offline.")
            await set_scene_state(SceneState.Data.StateEnum.Started, str(e))
            return

    try:
        await asyncio.gather(*[cleanup_object(obj) for obj in glob.SCENE_OBJECT_INSTANCES.values()])
    except Arcor2Exception as e:
        glob.logger.exception("Exception occurred while cleaning up objects.")
        await set_scene_state(SceneState.Data.StateEnum.Stopped, str(e))
    else:
        await set_scene_state(SceneState.Data.StateEnum.Stopped)

    glob.SCENE_OBJECT_INSTANCES.clear()


async def start_scene() -> None:
    """Creates instances of scene objects."""

    glob.logger.info("Starting the scene.")

    assert glob.SCENE

    await set_scene_state(SceneState.Data.StateEnum.Starting)

    try:
        await scene_srv.stop()
    except Arcor2Exception:
        await set_scene_state(SceneState.Data.StateEnum.Stopped, "Failed to prepare for start.")
        return

    object_overrides: Dict[str, List[Parameter]] = {}

    if glob.PROJECT:
        object_overrides = glob.PROJECT.overrides

    # object initialization could take some time - let's do it in parallel
    tasks = [
        asyncio.ensure_future(
            create_object_instance(obj, object_overrides[obj.id] if obj.id in object_overrides else None)
        )
        for obj in glob.SCENE.objects
    ]

    try:
        await asyncio.gather(*tasks)
    except Arcor2Exception as e:
        for t in tasks:
            t.cancel()
        glob.logger.exception("Failed to create instances.")
        await stop_scene(str(e))
        return

    try:
        await scene_srv.start()
    except Arcor2Exception as e:
        glob.logger.exception("Failed to go online.")
        await stop_scene(str(e))
        return

    await set_scene_state(SceneState.Data.StateEnum.Started)
    assert scene_started()
    assert await scene_srv.started()



from ast import AST
from dataclasses import dataclass, field
from typing import Dict, Optional, Type

from arcor2.object_types.abstract import Generic
from arcor2_arserver_data.objects import ObjectAction, ObjectTypeMeta
from arcor2_arserver_data.robot import RobotMeta


@dataclass
class ObjectTypeData:

    meta: ObjectTypeMeta
    type_def: Optional[Type[Generic]] = None
    actions: Dict[str, ObjectAction] = field(default_factory=dict)
    ast: Optional[AST] = None
    robot_meta: Optional[RobotMeta] = None

    def __post_init__(self) -> None:
        if not self.meta.disabled:
            assert self.type_def is not None
            assert self.ast is not None


ObjectTypeDict = Dict[str, ObjectTypeData]



import copy
import inspect
import os
from ast import AST
from typing import Dict, List, Optional, Type, get_type_hints

import humps
import typing_inspect
from dataclasses_jsonschema import JsonSchemaMixin

from arcor2.data.common import ActionMetadata
from arcor2.data.object_type import ParameterMeta
from arcor2.docstring import parse_docstring
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import Generic, GenericWithPose
from arcor2.object_types.utils import built_in_types, get_settings_def, iterate_over_actions
from arcor2.parameter_plugins import ParameterPluginException
from arcor2.parameter_plugins.utils import plugin_from_type
from arcor2.source.utils import SourceException, find_function, parse_def
from arcor2_arserver import globals as glob
from arcor2_arserver import settings
from arcor2_arserver.object_types.data import ObjectTypeData, ObjectTypeDict
from arcor2_arserver_data.objects import ObjectAction, ObjectTypeMeta


class ObjectTypeException(Arcor2Exception):
    pass


def remove_object_type(obj_type_id: str) -> None:

    path = os.path.join(settings.OBJECT_TYPE_PATH, settings.OBJECT_TYPE_MODULE, f"{humps.depascalize(obj_type_id)}.py")
    glob.logger.debug(f"Deleting {path}.")

    try:
        os.remove(path)
    except FileNotFoundError as e:
        raise Arcor2Exception(f"File for {obj_type_id} was not found.") from e


def obj_description_from_base(data: ObjectTypeDict, obj_type: ObjectTypeMeta) -> str:

    try:
        obj = data[obj_type.base]
    except KeyError:
        raise Arcor2Exception(f"Unknown object type: {obj_type}.")

    if obj.meta.description:
        return obj.meta.description

    if not obj.meta.base:
        return ""

    return obj_description_from_base(data, data[obj.meta.base].meta)


def get_dataclass_params(type_def: Type[JsonSchemaMixin]) -> List[ParameterMeta]:
    """Analyzes properties of dataclass and returns their metadata.

    :param type_def:
    :return:
    """

    ret: List[ParameterMeta] = []

    sig = inspect.signature(type_def.__init__)

    # TODO Will this work for inherited properties? Make a test! There is also dataclasses.fields maybe...
    # TODO use inspect.signature to get reliable order of parameters
    for name, ttype in get_type_hints(type_def.__init__).items():
        if name == "return":
            continue

        if issubclass(ttype, JsonSchemaMixin):
            pm = ParameterMeta(name=name, type="dataclass")  # TODO come-up with plugin for this?
            pm.children = get_dataclass_params(ttype)
        else:
            param_type = plugin_from_type(ttype)
            assert param_type is not None
            pm = ParameterMeta(name=name, type=param_type.type_name())

            def_val = sig.parameters[name].default
            if def_val is not inspect.Parameter.empty:
                pm.default_value = param_type.value_to_json(def_val)

            # TODO description, ranges, etc.

        ret.append(pm)

    return ret


def meta_from_def(type_def: Type[Generic], built_in: bool = False) -> ObjectTypeMeta:

    obj = ObjectTypeMeta(
        type_def.__name__,
        type_def.description(),
        built_in=built_in,
        abstract=type_def.abstract(),
        has_pose=issubclass(type_def, GenericWithPose),
    )

    for base in inspect.getmro(type_def)[1:-1]:  # skip type itself (first base) and the last one (object)
        if issubclass(base, Generic):
            obj.base = base.__name__
            break

    try:
        obj.settings = get_dataclass_params(get_settings_def(type_def))
    except Arcor2Exception:
        pass

    return obj


def built_in_types_data() -> ObjectTypeDict:

    ret: ObjectTypeDict = {}

    # built-in object types / services
    for _, type_def in built_in_types():

        assert issubclass(type_def, Generic)

        ast = parse_def(type_def)

        d = ObjectTypeData(meta_from_def(type_def, built_in=True), type_def, object_actions(type_def, ast), ast)

        ret[d.meta.type] = d

    return ret


class IgnoreActionException(Arcor2Exception):
    pass


def object_actions(type_def: Type[Generic], tree: AST) -> Dict[str, ObjectAction]:

    ret: Dict[str, ObjectAction] = {}

    # ...inspect.ismethod does not work on un-initialized classes
    for method_name, method_def in iterate_over_actions(type_def):

        meta: ActionMetadata = method_def.__action__  # type: ignore

        data = ObjectAction(name=method_name, meta=meta)

        if method_name in type_def.CANCEL_MAPPING:
            meta.cancellable = True

        try:

            if not method_def.__doc__:
                doc = {}
            else:
                doc = parse_docstring(method_def.__doc__)
                doc_short = doc["short_description"]
                if doc_short:
                    data.description = doc_short

            signature = inspect.signature(method_def)  # sig.parameters is OrderedDict

            try:
                method_tree = find_function(method_name, tree)
            except SourceException:
                # function is probably defined in predecessor, will be added later
                continue

            hints = get_type_hints(method_def)  # standard (unordered) dict

            if "an" not in signature.parameters.keys():
                raise IgnoreActionException("Action is missing 'an' parameter.")

            try:
                if hints["an"] != Optional[str]:
                    raise IgnoreActionException("Parameter 'an' has invalid type annotation.")
            except KeyError:
                raise IgnoreActionException("Parameter 'an' is missing type annotation.")

            parameter_names_without_self = list(signature.parameters.keys())[1:]

            if not parameter_names_without_self or parameter_names_without_self[-1] != "an":
                raise IgnoreActionException("The 'an' parameter have to be the last one.")

            # handle return
            try:
                return_ttype = hints["return"]
            except KeyError:
                raise IgnoreActionException("Action is missing return type annotation.")

            # ...just ignore NoneType for returns
            if return_ttype != type(None):  # noqa: E721

                if typing_inspect.is_tuple_type(return_ttype):
                    for arg in typing_inspect.get_args(return_ttype):
                        resolved_param = plugin_from_type(arg)
                        if resolved_param is None:
                            raise IgnoreActionException("None in return tuple is not supported.")
                        data.returns.append(resolved_param.type_name())
                else:
                    # TODO resolving needed for e.g. enums - add possible values to action metadata somewhere?
                    data.returns = [plugin_from_type(return_ttype).type_name()]

            for name in parameter_names_without_self[:-1]:  # omit also an

                try:
                    ttype = hints[name]
                except KeyError:
                    raise IgnoreActionException(f"Parameter {name} is missing type annotation.")

                param_type = plugin_from_type(ttype)

                assert param_type is not None

                args = ParameterMeta(name=name, type=param_type.type_name())
                try:
                    param_type.meta(args, method_def, method_tree)
                except ParameterPluginException as e:
                    raise IgnoreActionException(e) from e

                if name in type_def.DYNAMIC_PARAMS:
                    args.dynamic_value = True
                    dvp = type_def.DYNAMIC_PARAMS[name][1]
                    if dvp:
                        args.dynamic_value_parents = dvp

                def_val = signature.parameters[name].default
                if def_val is not inspect.Parameter.empty:
                    args.default_value = param_type.value_to_json(def_val)

                try:
                    args.description = doc["params"][name].strip()
                except KeyError:
                    pass

                data.parameters.append(args)

        except Arcor2Exception as e:
            data.disabled = True
            data.problem = str(e)
            glob.logger.warn(f"Disabling action {method_name} of  {type_def.__name__}. {str(e)}")

        ret[data.name] = data

    return ret


def add_ancestor_actions(obj_type_name: str, object_types: ObjectTypeDict) -> None:

    base_name = object_types[obj_type_name].meta.base

    if not base_name:
        return

    if object_types[base_name].meta.base:
        add_ancestor_actions(base_name, object_types)

    # do not add action from base if it is overridden in child
    # TODO rewrite (use adv. of dict!)
    for base_action in object_types[base_name].actions.values():
        for obj_action in object_types[obj_type_name].actions.values():
            if base_action.name == obj_action.name:

                # built-in object has no "origins" yet
                if not obj_action.origins:
                    obj_action.origins = base_name
                break
        else:
            action = copy.deepcopy(base_action)
            if not action.origins:
                action.origins = base_name
            object_types[obj_type_name].actions[action.name] = action


__all__ = [
    ObjectTypeException.__name__,
    ObjectTypeData.__name__,
    "ObjectTypeDict",
    meta_from_def.__name__,
    built_in_types_data.__name__,
    object_actions.__name__,
    add_ancestor_actions.__name__,
]



import asyncio
import os
from typing import List, Optional, Set, Type, Union

from arcor2 import helpers as hlp
from arcor2.clients import aio_persistent_storage as ps
from arcor2.data.events import Event
from arcor2.data.object_type import ObjectModel
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types import utils as otu
from arcor2.object_types.abstract import Generic, Robot
from arcor2.object_types.utils import built_in_types_names, get_containing_module_sources, prepare_object_types_dir
from arcor2.parameter_plugins.base import TypesDict
from arcor2.source.utils import parse
from arcor2_arserver import globals as glob
from arcor2_arserver import notifications as notif
from arcor2_arserver import settings
from arcor2_arserver.clients import persistent_storage as storage
from arcor2_arserver.object_types.utils import (
    ObjectTypeData,
    ObjectTypeDict,
    add_ancestor_actions,
    built_in_types_data,
    meta_from_def,
    obj_description_from_base,
    object_actions,
    remove_object_type,
)
from arcor2_arserver.robot import get_robot_meta
from arcor2_arserver_data.events.objects import ChangedObjectTypes
from arcor2_arserver_data.objects import ObjectTypeMeta


def get_types_dict() -> TypesDict:

    return {k: v.type_def for k, v in glob.OBJECT_TYPES.items() if v.type_def is not None}


def get_obj_type_name(object_id: str) -> str:

    assert glob.SCENE

    try:
        return glob.SCENE.object(object_id).type
    except KeyError:
        raise Arcor2Exception("Unknown object id.")


def get_obj_type_data(object_id: str) -> ObjectTypeData:

    try:
        return glob.OBJECT_TYPES[get_obj_type_name(object_id)]
    except KeyError:
        raise Arcor2Exception("Unknown object type.")


def valid_object_types() -> ObjectTypeDict:
    """To get only valid (not disabled) types.

    :return:
    """

    return {obj_type: obj for obj_type, obj in glob.OBJECT_TYPES.items() if not obj.meta.disabled}


async def handle_robot_urdf(robot: Type[Robot]) -> None:

    if not robot.urdf_package_name:
        return

    file_path = os.path.join(settings.URDF_PATH, robot.urdf_package_name)

    try:
        await ps.save_mesh_file(robot.urdf_package_name, file_path)
    except Arcor2Exception:
        glob.logger.exception(f"Failed to download URDF for {robot.__name__}.")


async def get_object_data(object_types: ObjectTypeDict, obj_id: str) -> None:

    glob.logger.debug(f"Processing {obj_id}.")

    if obj_id in object_types:
        glob.logger.debug(f"{obj_id} already processed, skipping...")
        return

    obj = await storage.get_object_type(obj_id)

    if obj_id in glob.OBJECT_TYPES and glob.OBJECT_TYPES[obj_id].type_def is not None:

        stored_type_def = glob.OBJECT_TYPES[obj_id].type_def
        assert stored_type_def
        if hash(get_containing_module_sources(stored_type_def)) == hash(obj.source):
            glob.logger.debug(f"No need to update {obj_id}.")
            return

    try:
        base = otu.base_from_source(obj.source, obj_id)
        if base and base not in object_types.keys() | built_in_types_names():
            glob.logger.debug(f"Getting base class {base} for {obj_id}.")
            await get_object_data(object_types, base)
    except Arcor2Exception:
        object_types[obj_id] = ObjectTypeData(
            ObjectTypeMeta(obj_id, "Object type disabled.", disabled=True, problem="Can't get base.")
        )
        return

    glob.logger.debug(f"Updating {obj_id}.")

    try:
        type_def = await hlp.run_in_executor(
            hlp.save_and_import_type_def,
            obj.source,
            obj.id,
            Generic,
            settings.OBJECT_TYPE_PATH,
            settings.OBJECT_TYPE_MODULE,
        )
        assert issubclass(type_def, Generic)
        meta = meta_from_def(type_def)
        otu.get_settings_def(type_def)  # just to check if settings are ok
    except Arcor2Exception as e:
        glob.logger.warning(f"Disabling object type {obj.id}.")
        glob.logger.debug(e, exc_info=True)
        object_types[obj_id] = ObjectTypeData(
            ObjectTypeMeta(obj_id, "Object type disabled.", disabled=True, problem=str(e))
        )
        return

    if obj.model:
        try:
            model = await storage.get_model(obj.model.id, obj.model.type)
        except Arcor2Exception:
            glob.logger.error(f"{obj.model.id}: failed to get collision model of type {obj.model.type}.")
            meta.disabled = True
            meta.problem = "Can't get collision model."
            object_types[obj_id] = ObjectTypeData(meta)
            return

        kwargs = {model.type().value.lower(): model}
        meta.object_model = ObjectModel(model.type(), **kwargs)  # type: ignore

    ast = parse(obj.source)
    otd = ObjectTypeData(meta, type_def, object_actions(type_def, ast), ast)

    object_types[obj_id] = otd


async def get_object_types() -> None:
    """Serves to initialize or update knowledge about awailable ObjectTypes.

    :return:
    """

    assert glob.SCENE is None

    initialization = False

    # initialize with built-in types, this has to be done just once
    if not glob.OBJECT_TYPES:
        glob.logger.debug("Initialization of object types.")
        initialization = True
        await hlp.run_in_executor(prepare_object_types_dir, settings.OBJECT_TYPE_PATH, settings.OBJECT_TYPE_MODULE)
        glob.OBJECT_TYPES.update(built_in_types_data())

    updated_object_types: ObjectTypeDict = {}

    object_type_ids: Union[Set[str], List[str]] = {it.id for it in (await storage.get_object_type_ids()).items}

    if __debug__:  # this should uncover potential problems with order in which ObjectTypes are processed
        import random

        object_type_ids = list(object_type_ids)
        random.shuffle(object_type_ids)

    for obj_id in object_type_ids:
        await get_object_data(updated_object_types, obj_id)

    removed_object_ids = {
        obj for obj in glob.OBJECT_TYPES.keys() if obj not in object_type_ids
    } - built_in_types_names()
    updated_object_ids = {k for k in updated_object_types.keys() if k in glob.OBJECT_TYPES}
    new_object_ids = {k for k in updated_object_types.keys() if k not in glob.OBJECT_TYPES}

    glob.logger.debug(f"Removed ids: {removed_object_ids}")
    glob.logger.debug(f"Updated ids: {updated_object_ids}")
    glob.logger.debug(f"New ids: {new_object_ids}")

    if not initialization and removed_object_ids:

        # TODO remove it from sys.modules

        remove_evt = ChangedObjectTypes([v.meta for k, v in glob.OBJECT_TYPES.items() if k in removed_object_ids])
        remove_evt.change_type = Event.Type.REMOVE
        asyncio.ensure_future(notif.broadcast_event(remove_evt))

        for removed in removed_object_ids:
            assert removed not in built_in_types_names(), "Attempt to remove built-in type."
            del glob.OBJECT_TYPES[removed]
            await hlp.run_in_executor(remove_object_type, removed)

    glob.OBJECT_TYPES.update(updated_object_types)

    glob.logger.debug(f"All known ids: {glob.OBJECT_TYPES.keys()}")

    for obj_type in updated_object_types.values():

        # if description is missing, try to get it from ancestor(s)
        if not obj_type.meta.description:

            try:
                obj_type.meta.description = obj_description_from_base(glob.OBJECT_TYPES, obj_type.meta)
            except otu.DataError as e:
                glob.logger.error(f"Failed to get info from base for {obj_type}, error: '{e}'.")

        if not obj_type.meta.disabled and not obj_type.meta.built_in:
            add_ancestor_actions(obj_type.meta.type, glob.OBJECT_TYPES)

    if not initialization:

        if updated_object_ids:
            update_evt = ChangedObjectTypes([v.meta for k, v in glob.OBJECT_TYPES.items() if k in updated_object_ids])
            update_evt.change_type = Event.Type.UPDATE
            asyncio.ensure_future(notif.broadcast_event(update_evt))

        if new_object_ids:
            add_evt = ChangedObjectTypes([v.meta for k, v in glob.OBJECT_TYPES.items() if k in new_object_ids])
            add_evt.change_type = Event.Type.ADD
            asyncio.ensure_future(notif.broadcast_event(add_evt))

    for obj_type in updated_object_types.values():

        if obj_type.type_def and issubclass(obj_type.type_def, Robot) and not obj_type.type_def.abstract():
            await get_robot_meta(obj_type)
            asyncio.ensure_future(handle_robot_urdf(obj_type.type_def))

    # if object does not change but its base has changed, it has to be reloaded
    for obj_id, obj in glob.OBJECT_TYPES.items():

        if obj_id in updated_object_ids:
            continue

        if obj.type_def and obj.meta.base in updated_object_ids:

            glob.logger.debug(f"Re-importing {obj.meta.type} because its base {obj.meta.base} type has changed.")
            obj.type_def = await hlp.run_in_executor(
                hlp.import_type_def,
                obj.meta.type,
                Generic,
                settings.OBJECT_TYPE_PATH,
                settings.OBJECT_TYPE_MODULE,
            )


async def get_robot_instance(robot_id: str, end_effector_id: Optional[str] = None) -> Robot:

    if robot_id not in glob.SCENE_OBJECT_INSTANCES:
        raise Arcor2Exception("Robot not found.")

    robot_inst = glob.SCENE_OBJECT_INSTANCES[robot_id]
    if not isinstance(robot_inst, Robot):
        raise Arcor2Exception("Not a robot.")
    if end_effector_id and end_effector_id not in await hlp.run_in_executor(robot_inst.get_end_effectors_ids):
        raise Arcor2Exception("Unknown end effector ID.")
    return robot_inst



import inspect
from ast import AST
from typing import List, Optional, Set, Type

import arcor2.helpers as hlp
from arcor2.data import common
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import Robot
from arcor2_arserver import globals as glob
from arcor2_arserver import notifications as notif
from arcor2_arserver import objects_actions as osa
from arcor2_arserver.object_types.data import ObjectTypeData
from arcor2_arserver.object_types.source import function_implemented
from arcor2_arserver_data import events as sevts
from arcor2_arserver_data.robot import RobotMeta


class RobotPoseException(Arcor2Exception):
    pass


async def get_end_effectors(robot_id: str) -> Set[str]:
    """
    :param robot_id:
    :return: IDs of existing end effectors.
    """

    robot_inst = await osa.get_robot_instance(robot_id)
    return await hlp.run_in_executor(robot_inst.get_end_effectors_ids)


async def get_grippers(robot_id: str) -> Set[str]:
    """
    :param robot_id:
    :return: IDs of existing grippers.
    """

    robot_inst = await osa.get_robot_instance(robot_id)
    return await hlp.run_in_executor(robot_inst.grippers)


async def get_suctions(robot_id: str) -> Set[str]:
    """
    :param robot_id:
    :return: IDs of existing suctions.
    """

    robot_inst = await osa.get_robot_instance(robot_id)
    return await hlp.run_in_executor(robot_inst.suctions)


async def get_end_effector_pose(robot_id: str, end_effector: str) -> common.Pose:
    """
    :param robot_id:
    :param end_effector:
    :return: Global pose
    """

    robot_inst = await osa.get_robot_instance(robot_id, end_effector)
    return await hlp.run_in_executor(robot_inst.get_end_effector_pose, end_effector)


async def get_robot_joints(robot_id: str) -> List[common.Joint]:
    """
    :param robot_id:
    :return: List of joints
    """

    robot_inst = await osa.get_robot_instance(robot_id)
    return await hlp.run_in_executor(robot_inst.robot_joints)


def feature(tree: AST, robot_type: Type[Robot], func_name: str) -> bool:

    if not function_implemented(tree, func_name):
        glob.logger.debug(f"robot_type {robot_type.__name__}, func_name {func_name} not implemented.")
        return False

    sign = inspect.signature(getattr(robot_type, func_name))
    res = inspect.signature(getattr(Robot, func_name)) == sign
    if not res:
        glob.logger.debug(f"robot_type {robot_type.__name__}, func_name {func_name} has invalid signature.")
    return res


def _feature(type_def: Type[Robot], method_name: str) -> bool:

    assert glob.OBJECT_TYPES

    method = getattr(type_def, method_name)
    where_it_is_defined = glob.OBJECT_TYPES[method.__qualname__.split(".")[0]]

    if where_it_is_defined.type_def is Robot or where_it_is_defined.meta.disabled:
        return False

    assert where_it_is_defined.ast is not None
    assert where_it_is_defined.type_def is not None
    assert issubclass(where_it_is_defined.type_def, Robot)

    return feature(where_it_is_defined.ast, where_it_is_defined.type_def, method_name)


async def get_robot_meta(obj_type: ObjectTypeData) -> None:

    if obj_type.meta.disabled:
        raise Arcor2Exception("Disabled object type.")

    assert obj_type.type_def is not None

    if not issubclass(obj_type.type_def, Robot):
        raise Arcor2Exception("Not a robot.")

    obj_type.robot_meta = RobotMeta(obj_type.meta.type, obj_type.type_def.robot_type)

    # TODO automate this somehow
    obj_type.robot_meta.features.move_to_pose = _feature(obj_type.type_def, Robot.move_to_pose.__name__)
    obj_type.robot_meta.features.move_to_joints = _feature(obj_type.type_def, Robot.move_to_joints.__name__)
    obj_type.robot_meta.features.stop = _feature(obj_type.type_def, Robot.stop.__name__)
    obj_type.robot_meta.features.inverse_kinematics = _feature(obj_type.type_def, Robot.inverse_kinematics.__name__)
    obj_type.robot_meta.features.forward_kinematics = _feature(obj_type.type_def, Robot.forward_kinematics.__name__)
    obj_type.robot_meta.features.hand_teaching = _feature(obj_type.type_def, Robot.set_hand_teaching_mode.__name__)

    if issubclass(obj_type.type_def, Robot) and obj_type.type_def.urdf_package_name:
        obj_type.robot_meta.urdf_package_filename = obj_type.type_def.urdf_package_name


async def stop(robot_id: str) -> None:

    robot_inst = await osa.get_robot_instance(robot_id)

    if not robot_inst.move_in_progress:
        raise Arcor2Exception("Robot is not moving.")

    await hlp.run_in_executor(robot_inst.stop)


async def ik(
    robot_id: str,
    end_effector_id: str,
    pose: common.Pose,
    start_joints: Optional[List[common.Joint]] = None,
    avoid_collisions: bool = True,
) -> List[common.Joint]:

    robot_inst = await osa.get_robot_instance(robot_id)

    return await hlp.run_in_executor(
        robot_inst.inverse_kinematics, end_effector_id, pose, start_joints, avoid_collisions
    )


async def fk(robot_id: str, end_effector_id: str, joints: List[common.Joint]) -> common.Pose:

    robot_inst = await osa.get_robot_instance(robot_id)
    return await hlp.run_in_executor(robot_inst.forward_kinematics, end_effector_id, joints)


async def check_reachability(robot_id: str, end_effector_id: str, pose: common.Pose, safe: bool = True) -> None:

    otd = osa.get_obj_type_data(robot_id)
    if otd.robot_meta and otd.robot_meta.features.inverse_kinematics:
        try:
            await ik(robot_id, end_effector_id, pose, avoid_collisions=safe)
        except Arcor2Exception as e:
            raise Arcor2Exception("Unreachable pose.") from e


async def check_robot_before_move(robot_id: str) -> None:

    robot_inst = await osa.get_robot_instance(robot_id)
    robot_inst.check_if_ready_to_move()

    if glob.RUNNING_ACTION:

        assert glob.PROJECT
        action = glob.PROJECT.action(glob.RUNNING_ACTION)
        obj_id_str, _ = action.parse_type()

        if robot_id == obj_id_str:
            raise Arcor2Exception("Robot is executing action.")


async def _move_to_pose(robot_id: str, end_effector_id: str, pose: common.Pose, speed: float, safe: bool) -> None:
    # TODO newly connected interface should be notified somehow (general solution for such cases would be great!)

    robot_inst = await osa.get_robot_instance(robot_id, end_effector_id)

    try:
        await hlp.run_in_executor(robot_inst.move_to_pose, end_effector_id, pose, speed, safe)
    except Arcor2Exception as e:
        glob.logger.error(f"Robot movement failed with: {str(e)}")
        raise


async def move_to_pose(robot_id: str, end_effector_id: str, pose: common.Pose, speed: float, safe: bool) -> None:

    Data = sevts.r.RobotMoveToPose.Data

    await notif.broadcast_event(
        sevts.r.RobotMoveToPose(Data(Data.MoveEventType.START, robot_id, end_effector_id, pose, safe))
    )

    try:
        await _move_to_pose(robot_id, end_effector_id, pose, speed, safe)

    except Arcor2Exception as e:
        await notif.broadcast_event(
            sevts.r.RobotMoveToPose(
                Data(Data.MoveEventType.FAILED, robot_id, end_effector_id, pose, safe, message=str(e))
            )
        )
        return

    await notif.broadcast_event(
        sevts.r.RobotMoveToPose(Data(Data.MoveEventType.END, robot_id, end_effector_id, pose, safe))
    )


async def move_to_ap_orientation(
    robot_id: str, end_effector_id: str, pose: common.Pose, speed: float, orientation_id: str, safe: bool
) -> None:

    Data = sevts.r.RobotMoveToActionPointOrientation.Data

    await notif.broadcast_event(
        sevts.r.RobotMoveToActionPointOrientation(
            Data(Data.MoveEventType.START, robot_id, end_effector_id, orientation_id, safe)
        )
    )

    try:
        await _move_to_pose(robot_id, end_effector_id, pose, speed, safe)

    except Arcor2Exception as e:
        await notif.broadcast_event(
            sevts.r.RobotMoveToActionPointOrientation(
                Data(Data.MoveEventType.FAILED, robot_id, end_effector_id, orientation_id, safe, message=str(e))
            )
        )
        return

    await notif.broadcast_event(
        sevts.r.RobotMoveToActionPointOrientation(
            Data(Data.MoveEventType.END, robot_id, end_effector_id, orientation_id, safe)
        )
    )


async def _move_to_joints(robot_id: str, joints: List[common.Joint], speed: float, safe: bool) -> None:

    # TODO newly connected interface should be notified somehow (general solution for such cases would be great!)

    robot_inst = await osa.get_robot_instance(robot_id)

    try:
        await hlp.run_in_executor(robot_inst.move_to_joints, joints, speed, safe)
    except Arcor2Exception as e:
        glob.logger.error(f"Robot movement failed with: {str(e)}")
        raise


async def move_to_joints(robot_id: str, joints: List[common.Joint], speed: float, safe: bool) -> None:

    Data = sevts.r.RobotMoveToJoints.Data

    await notif.broadcast_event(sevts.r.RobotMoveToJoints(Data(Data.MoveEventType.START, robot_id, joints, safe)))

    try:

        await _move_to_joints(robot_id, joints, speed, safe)

    except Arcor2Exception as e:

        await notif.broadcast_event(
            sevts.r.RobotMoveToJoints(Data(Data.MoveEventType.FAILED, robot_id, joints, safe, message=str(e)))
        )

        return

    await notif.broadcast_event(sevts.r.RobotMoveToJoints(Data(Data.MoveEventType.END, robot_id, joints, safe)))


async def move_to_ap_joints(
    robot_id: str, joints: List[common.Joint], speed: float, joints_id: str, safe: bool
) -> None:

    Data = sevts.r.RobotMoveToActionPointJoints.Data

    await notif.broadcast_event(
        sevts.r.RobotMoveToActionPointJoints(Data(Data.MoveEventType.START, robot_id, joints_id, safe))
    )

    try:

        await _move_to_joints(robot_id, joints, speed, safe)

    except Arcor2Exception as e:

        await notif.broadcast_event(
            sevts.r.RobotMoveToActionPointJoints(
                Data(Data.MoveEventType.FAILED, robot_id, joints_id, safe, message=str(e))
            )
        )

        return

    await notif.broadcast_event(
        sevts.r.RobotMoveToActionPointJoints(Data(Data.MoveEventType.END, robot_id, joints_id, safe))
    )



import asyncio
from typing import Dict, List

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import helpers as hlp
from arcor2.cached import CachedScene
from arcor2.clients import aio_scene_service as scene_srv
from arcor2.data import events, rpc
from arcor2.data.common import Parameter, Pose, Position, SceneObject
from arcor2.data.object_type import Model3dType
from arcor2.data.scene import MeshFocusAction
from arcor2.exceptions import Arcor2Exception
from arcor2.object_types.abstract import GenericWithPose, Robot
from arcor2.source.utils import tree_to_str
from arcor2_arserver import globals as glob
from arcor2_arserver import notifications as notif
from arcor2_arserver import objects_actions as osa
from arcor2_arserver import settings
from arcor2_arserver.clients import persistent_storage as storage
from arcor2_arserver.decorators import no_project, project_needed, scene_needed
from arcor2_arserver.object_types.data import ObjectTypeData
from arcor2_arserver.object_types.source import new_object_type
from arcor2_arserver.object_types.utils import add_ancestor_actions, object_actions, remove_object_type
from arcor2_arserver.robot import get_end_effector_pose
from arcor2_arserver.scene import ensure_scene_started, scenes, update_scene_object_pose
from arcor2_arserver_data import events as sevts
from arcor2_arserver_data import rpc as srpc

FOCUS_OBJECT: Dict[str, Dict[int, Pose]] = {}  # object_id / idx, pose
FOCUS_OBJECT_ROBOT: Dict[str, rpc.common.RobotArg] = {}  # key: object_id


def clean_up_after_focus(obj_id: str) -> None:

    try:
        del FOCUS_OBJECT[obj_id]
    except KeyError:
        pass

    try:
        del FOCUS_OBJECT_ROBOT[obj_id]
    except KeyError:
        pass


@scene_needed
@no_project
async def focus_object_start_cb(req: srpc.o.FocusObjectStart.Request, ui: WsClient) -> None:

    ensure_scene_started()

    obj_id = req.args.object_id

    if obj_id in FOCUS_OBJECT_ROBOT:
        raise Arcor2Exception("Focusing already started.")

    if obj_id not in glob.SCENE_OBJECT_INSTANCES:
        raise Arcor2Exception("Unknown object.")

    inst = await osa.get_robot_instance(req.args.robot.robot_id, req.args.robot.end_effector)

    robot_type = glob.OBJECT_TYPES[inst.__class__.__name__]
    assert robot_type.robot_meta

    obj_type = glob.OBJECT_TYPES[osa.get_obj_type_name(obj_id)].meta

    if not obj_type.has_pose:
        raise Arcor2Exception("Only available for objects with pose.")

    if not obj_type.object_model or obj_type.object_model.type != Model3dType.MESH:
        raise Arcor2Exception("Only available for objects with mesh model.")

    assert obj_type.object_model.mesh

    focus_points = obj_type.object_model.mesh.focus_points

    if not focus_points:
        raise Arcor2Exception("focusPoints not defined for the mesh.")

    FOCUS_OBJECT_ROBOT[req.args.object_id] = req.args.robot
    FOCUS_OBJECT[obj_id] = {}
    glob.logger.info(f"Start of focusing for {obj_id}.")
    return None


@no_project
async def focus_object_cb(req: srpc.o.FocusObject.Request, ui: WsClient) -> srpc.o.FocusObject.Response:

    obj_id = req.args.object_id
    pt_idx = req.args.point_idx

    if obj_id not in glob.SCENE_OBJECT_INSTANCES:
        raise Arcor2Exception("Unknown object_id.")

    obj_type = glob.OBJECT_TYPES[osa.get_obj_type_name(obj_id)].meta

    assert obj_type.has_pose
    assert obj_type.object_model
    assert obj_type.object_model.mesh

    focus_points = obj_type.object_model.mesh.focus_points

    assert focus_points

    if pt_idx < 0 or pt_idx > len(focus_points) - 1:
        raise Arcor2Exception("Index out of range.")

    if obj_id not in FOCUS_OBJECT:
        glob.logger.info(f"Start of focusing for {obj_id}.")
        FOCUS_OBJECT[obj_id] = {}

    robot_id, end_effector = FOCUS_OBJECT_ROBOT[obj_id].as_tuple()

    FOCUS_OBJECT[obj_id][pt_idx] = await get_end_effector_pose(robot_id, end_effector)

    r = srpc.o.FocusObject.Response()
    r.data = r.Data(finished_indexes=list(FOCUS_OBJECT[obj_id].keys()))
    return r


@scene_needed
@no_project
async def focus_object_done_cb(req: srpc.o.FocusObjectDone.Request, ui: WsClient) -> None:

    obj_id = req.args.id

    if obj_id not in FOCUS_OBJECT:
        raise Arcor2Exception("focusObjectStart/focusObject has to be called first.")

    obj_type = glob.OBJECT_TYPES[osa.get_obj_type_name(obj_id)].meta

    assert obj_type.object_model
    assert obj_type.object_model.mesh

    focus_points = obj_type.object_model.mesh.focus_points

    assert focus_points

    if len(FOCUS_OBJECT[obj_id]) < len(focus_points):
        raise Arcor2Exception("Not all points were done.")

    assert glob.SCENE

    obj = glob.SCENE.object(obj_id)
    assert obj.pose

    obj_inst = glob.SCENE_OBJECT_INSTANCES[obj_id]
    assert isinstance(obj_inst, GenericWithPose)

    fp: List[Position] = []
    rp: List[Position] = []

    for idx, pose in FOCUS_OBJECT[obj_id].items():

        fp.append(focus_points[idx].position)
        rp.append(pose.position)

    mfa = MeshFocusAction(fp, rp)

    glob.logger.debug(f"Attempt to focus for object {obj_id}, data: {mfa}")

    try:
        new_pose = await scene_srv.focus(mfa)
    except scene_srv.SceneServiceException as e:
        glob.logger.error(f"Focus failed with: {e}, mfa: {mfa}.")
        raise Arcor2Exception("Focusing failed.") from e

    glob.logger.info(f"Done focusing for {obj_id}.")

    clean_up_after_focus(obj_id)

    asyncio.ensure_future(update_scene_object_pose(obj, new_pose, obj_inst))

    return None


async def new_object_type_cb(req: srpc.o.NewObjectType.Request, ui: WsClient) -> None:

    meta = req.args

    if meta.type in glob.OBJECT_TYPES:
        raise Arcor2Exception("Object type already exists.")

    hlp.is_valid_type(meta.type)

    if meta.base not in glob.OBJECT_TYPES:
        raise Arcor2Exception(
            f"Unknown base object type '{meta.base}', " f"known types are: {', '.join(glob.OBJECT_TYPES.keys())}."
        )

    base = glob.OBJECT_TYPES[meta.base]

    if base.meta.disabled:
        raise Arcor2Exception("Base object is disabled.")

    assert base.type_def is not None

    if issubclass(base.type_def, Robot):
        raise Arcor2Exception("Can't subclass Robot.")

    meta.has_pose = issubclass(base.type_def, GenericWithPose)

    if not meta.has_pose and meta.object_model:
        raise Arcor2Exception("Object without pose can't have collision model.")

    if req.dry_run:
        return None

    obj = meta.to_object_type()
    ast = new_object_type(glob.OBJECT_TYPES[meta.base].meta, meta)
    obj.source = tree_to_str(ast)

    if meta.object_model:

        if meta.object_model.type == Model3dType.MESH:

            # TODO check whether mesh id exists - if so, then use existing mesh, if not, upload a new one
            # ...get whole mesh (focus_points) based on mesh id
            assert meta.object_model.mesh
            try:
                meta.object_model.mesh = await storage.get_mesh(meta.object_model.mesh.id)
            except storage.ProjectServiceException as e:
                glob.logger.error(e)
                raise Arcor2Exception(f"Mesh ID {meta.object_model.mesh.id} does not exist.")

        else:

            meta.object_model.model().id = meta.type
            await storage.put_model(meta.object_model.model())

    type_def = await hlp.run_in_executor(
        hlp.save_and_import_type_def,
        obj.source,
        obj.id,
        base.type_def,
        settings.OBJECT_TYPE_PATH,
        settings.OBJECT_TYPE_MODULE,
    )
    assert issubclass(type_def, base.type_def)
    actions = object_actions(type_def, ast)

    await storage.update_object_type(obj)

    glob.OBJECT_TYPES[meta.type] = ObjectTypeData(meta, type_def, actions, ast)
    add_ancestor_actions(meta.type, glob.OBJECT_TYPES)

    evt = sevts.o.ChangedObjectTypes([meta])
    evt.change_type = events.Event.Type.ADD
    asyncio.ensure_future(notif.broadcast_event(evt))
    return None


async def get_object_actions_cb(req: srpc.o.GetActions.Request, ui: WsClient) -> srpc.o.GetActions.Response:

    try:
        return srpc.o.GetActions.Response(data=list(glob.OBJECT_TYPES[req.args.type].actions.values()))
    except KeyError:
        raise Arcor2Exception(f"Unknown object type: '{req.args.type}'.")


async def get_object_types_cb(req: srpc.o.GetObjectTypes.Request, ui: WsClient) -> srpc.o.GetObjectTypes.Response:
    return srpc.o.GetObjectTypes.Response(data=[obj.meta for obj in glob.OBJECT_TYPES.values()])


def check_scene_for_object_type(scene: CachedScene, object_type: str) -> None:

    if list(scene.objects_of_type(object_type)):
        raise Arcor2Exception(f"Object type used in scene '{scene.name}'.")


async def delete_object_type_cb(req: srpc.o.DeleteObjectType.Request, ui: WsClient) -> None:

    try:
        obj_type = glob.OBJECT_TYPES[req.args.id]
    except KeyError:
        raise Arcor2Exception("Unknown object type.")

    if obj_type.meta.built_in:
        raise Arcor2Exception("Can't delete built-in type.")

    for obj in glob.OBJECT_TYPES.values():
        if obj.meta.base == req.args.id:
            raise Arcor2Exception(f"Object type is base of '{obj.meta.type}'.")

    async for scene in scenes():
        check_scene_for_object_type(scene, req.args.id)

    if glob.SCENE:
        check_scene_for_object_type(glob.SCENE, req.args.id)

    if req.dry_run:
        return

    await storage.delete_object_type(req.args.id)

    # do not care so much if delete_model fails
    if obj_type.meta.object_model:
        try:
            await storage.delete_model(obj_type.meta.object_model.model().id)
        except storage.ProjectServiceException as e:
            glob.logger.error(str(e))

    del glob.OBJECT_TYPES[req.args.id]
    remove_object_type(req.args.id)

    evt = sevts.o.ChangedObjectTypes([obj_type.meta])
    evt.change_type = events.Event.Type.REMOVE
    asyncio.ensure_future(notif.broadcast_event(evt))


def check_override(obj_id: str, override: Parameter, add_new_one: bool = False) -> SceneObject:

    assert glob.SCENE
    assert glob.PROJECT

    obj = glob.SCENE.object(obj_id)

    for par in glob.OBJECT_TYPES[obj.type].meta.settings:
        if par.name == override.name:
            if par.type != override.type:
                raise Arcor2Exception("Override can't change parameter type.")
            break
    else:
        raise Arcor2Exception("Unknown parameter name.")

    if add_new_one:
        try:
            for existing_override in glob.PROJECT.overrides[obj.id]:
                if override.name == existing_override.name:
                    raise Arcor2Exception("Override already exists.")
        except KeyError:
            pass
    else:
        if obj.id not in glob.PROJECT.overrides:
            raise Arcor2Exception("There are no overrides for the object.")

        for override in glob.PROJECT.overrides[obj.id]:
            if override.name == override.name:
                break
        else:
            raise Arcor2Exception("Override not found.")

    return obj


@project_needed
async def add_override_cb(req: srpc.o.AddOverride.Request, ui: WsClient) -> None:

    assert glob.PROJECT

    obj = check_override(req.args.id, req.args.override, add_new_one=True)

    if req.dry_run:
        return

    if obj.id not in glob.PROJECT.overrides:
        glob.PROJECT.overrides[obj.id] = []

    glob.PROJECT.overrides[obj.id].append(req.args.override)
    glob.PROJECT.update_modified()

    evt = sevts.o.OverrideUpdated(req.args.override)
    evt.change_type = events.Event.Type.ADD
    evt.parent_id = req.args.id
    asyncio.ensure_future(notif.broadcast_event(evt))


@project_needed
async def update_override_cb(req: srpc.o.UpdateOverride.Request, ui: WsClient) -> None:

    assert glob.PROJECT

    obj = check_override(req.args.id, req.args.override)

    if req.dry_run:
        return

    for override in glob.PROJECT.overrides[obj.id]:
        if override.name == override.name:
            override.value = req.args.override.value
    glob.PROJECT.update_modified()

    evt = sevts.o.OverrideUpdated(req.args.override)
    evt.change_type = events.Event.Type.UPDATE
    evt.parent_id = req.args.id
    asyncio.ensure_future(notif.broadcast_event(evt))


@project_needed
async def delete_override_cb(req: srpc.o.DeleteOverride.Request, ui: WsClient) -> None:

    assert glob.PROJECT

    obj = check_override(req.args.id, req.args.override)

    if req.dry_run:
        return

    glob.PROJECT.overrides[obj.id] = [ov for ov in glob.PROJECT.overrides[obj.id] if ov.name != req.args.override.name]

    if not glob.PROJECT.overrides[obj.id]:
        del glob.PROJECT.overrides[obj.id]

    glob.PROJECT.update_modified()

    evt = sevts.o.OverrideUpdated(req.args.override)
    evt.change_type = events.Event.Type.REMOVE
    evt.parent_id = req.args.id
    asyncio.ensure_future(notif.broadcast_event(evt))



import asyncio
import socket
from typing import Optional, Text, Tuple, Union

from arcor2.data.common import BroadcastInfo
from arcor2_arserver.globals import PORT

BROADCAST_PORT: int = 6006


def get_ip() -> str:

    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(("10.255.255.255", 1))
        ip = s.getsockname()[0]
    except socket.error:
        ip = "127.0.0.1"
    finally:
        s.close()
    return ip


def get_broadcast() -> str:

    ip = get_ip()
    ip_arr = ip.split(".")
    ip_arr[-1] = "255"
    return ".".join(ip_arr)


Address = Tuple[str, int]


class BroadcastProtocol(asyncio.DatagramProtocol):
    def __init__(self, target: Address, *, loop: Optional[asyncio.AbstractEventLoop] = None):
        self.target = target
        self.loop = asyncio.get_event_loop() if loop is None else loop

    def connection_made(self, transport):
        self.transport = transport
        sock: socket.socket = transport.get_extra_info("socket")
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        self.broadcast()

    def datagram_received(self, data: Union[bytes, Text], addr: Address):
        pass

    def broadcast(self) -> None:
        self.transport.sendto(BroadcastInfo(socket.gethostname(), PORT).to_json().encode(), self.target)
        self.loop.call_later(1, self.broadcast)


def main() -> None:

    loop = asyncio.get_event_loop()

    broadcast_coro = loop.create_datagram_endpoint(
        lambda: BroadcastProtocol((get_broadcast(), BROADCAST_PORT), loop=loop), local_addr=("0.0.0.0", BROADCAST_PORT)
    )

    loop.run_until_complete(broadcast_coro)
    loop.run_forever()
    loop.close()


if __name__ == "__main__":
    main()



import asyncio
from typing import Optional

from websockets.server import WebSocketServerProtocol

from arcor2 import ws_server
from arcor2.data import events
from arcor2_arserver import globals as glob


async def broadcast_event(event: events.Event, exclude_ui: Optional[WebSocketServerProtocol] = None) -> None:

    if (exclude_ui is None and glob.INTERFACES) or (exclude_ui and len(glob.INTERFACES) > 1):
        message = event.to_json()
        await asyncio.gather(
            *[ws_server.send_json_to_client(intf, message) for intf in glob.INTERFACES if intf != exclude_ui]
        )


async def event(interface: WebSocketServerProtocol, event: events.Event) -> None:
    await ws_server.send_json_to_client(interface, event.to_json())



import asyncio

from websockets.server import WebSocketServerProtocol as WsClient

from arcor2.exceptions import Arcor2Exception
from arcor2_arserver import decorators
from arcor2_arserver import globals as glob
from arcor2_arserver.execution import build_and_upload_package, run_temp_package
from arcor2_arserver_data import rpc


async def build_project_cb(req: rpc.b.BuildProject.Request, ui: WsClient) -> rpc.b.BuildProject.Response:
    """Builds project and uploads resulting package to the execution unit.

    :param req:
    :param ui:
    :return:
    """

    package_id = await build_and_upload_package(req.args.project_id, req.args.package_name)

    resp = rpc.b.BuildProject.Response()
    resp.data = resp.Data(package_id)
    return resp


@decorators.project_needed
async def temporary_package_cb(req: rpc.b.TemporaryPackage.Request, ui: WsClient) -> None:

    assert glob.PROJECT

    if glob.PROJECT.has_changes:
        raise Arcor2Exception("Project has unsaved changes.")

    package_id = await build_and_upload_package(
        glob.PROJECT.id, f"Temporary package for project '{glob.PROJECT.name}'."
    )

    asyncio.ensure_future(run_temp_package(package_id))
    return None



import asyncio
from typing import Any, AsyncIterator, Callable, Dict, List, Set, Tuple, Union

from arcor2 import helpers as hlp
from arcor2.action import results_to_json
from arcor2.cached import CachedProject, CachedScene, UpdateableCachedProject
from arcor2.data import common
from arcor2.exceptions import Arcor2Exception
from arcor2.parameter_plugins import ParameterPluginException
from arcor2.parameter_plugins.utils import known_parameter_types, plugin_from_type_name
from arcor2_arserver import globals as glob
from arcor2_arserver import notifications as notif
from arcor2_arserver.clients import persistent_storage as storage
from arcor2_arserver.objects_actions import get_types_dict
from arcor2_arserver.scene import open_scene
from arcor2_arserver_data.events.actions import ActionExecution, ActionResult
from arcor2_arserver_data.events.common import ShowMainScreen
from arcor2_arserver_data.events.project import ProjectClosed
from arcor2_arserver_data.objects import ObjectAction

PREV_RESULTS: Dict[str, Union[Tuple[Any], Any]] = {}


def remove_prev_result(action_id: str) -> None:

    try:
        del PREV_RESULTS[action_id]
    except KeyError:
        pass


async def notify_project_closed(project_id: str) -> None:

    proj_list = ShowMainScreen.Data.WhatEnum.ProjectsList

    await notif.broadcast_event(ProjectClosed())
    glob.MAIN_SCREEN = ShowMainScreen.Data(proj_list)
    await notif.broadcast_event(ShowMainScreen(ShowMainScreen.Data(proj_list, project_id)))


async def close_project() -> None:

    assert glob.PROJECT

    project_id = glob.PROJECT.project.id
    glob.SCENE = None
    glob.PROJECT = None
    PREV_RESULTS.clear()
    asyncio.ensure_future(notify_project_closed(project_id))


async def execute_action(action_method: Callable, params: List[Any]) -> None:

    assert glob.RUNNING_ACTION

    await notif.broadcast_event(ActionExecution(ActionExecution.Data(glob.RUNNING_ACTION)))

    evt = ActionResult(ActionResult.Data(glob.RUNNING_ACTION))

    try:
        action_result = await hlp.run_in_executor(action_method, *params)
    except (Arcor2Exception, AttributeError, TypeError) as e:
        glob.logger.error(f"Failed to run method {action_method.__name__} with params {params}. {str(e)}")
        glob.logger.debug(str(e), exc_info=True)
        evt.data.error = str(e)
    else:
        if action_result is not None:
            PREV_RESULTS[glob.RUNNING_ACTION] = action_result

        try:
            evt.data.results = results_to_json(action_result)
        except Arcor2Exception:
            glob.logger.error(f"Method {action_method.__name__} returned unsupported type of result: {action_result}.")

    if glob.RUNNING_ACTION is None:
        # action was cancelled, do not send any event
        return  # type: ignore  # action could be cancelled during its execution

    await notif.broadcast_event(evt)
    glob.RUNNING_ACTION = None
    glob.RUNNING_ACTION_PARAMS = None


def check_action_params(
    scene: CachedScene, project: CachedProject, action: common.Action, object_action: ObjectAction
) -> None:

    _, action_type = action.parse_type()

    assert action_type == object_action.name

    if len(object_action.parameters) != len(action.parameters):
        raise Arcor2Exception("Unexpected number of parameters.")

    for req_param in object_action.parameters:

        param = action.parameter(req_param.name)

        if param.type == common.ActionParameter.TypeEnum.CONSTANT:

            const = project.constant(param.str_from_value())

            param_meta = object_action.parameter(param.name)
            if param_meta.type != const.type:
                raise Arcor2Exception("Param type does not match constant type.")

        elif param.type == common.ActionParameter.TypeEnum.LINK:

            parsed_link = param.parse_link()

            if parsed_link.action_id == action.id:
                raise Arcor2Exception("Can't use own result as a parameter.")

            outputs = project.action(parsed_link.action_id).flow(parsed_link.flow_name).outputs

            assert len(outputs) == len(object_action.returns)

            param_meta = object_action.parameter(param.name)
            if param_meta.type != object_action.returns[parsed_link.output_index]:
                raise Arcor2Exception("Param type does not match action output type.")

        else:

            if param.type not in known_parameter_types():
                raise Arcor2Exception(f"Parameter {param.name} of action {action.name} has unknown type: {param.type}.")

            try:
                plugin_from_type_name(param.type).parameter_value(
                    get_types_dict(), scene, project, action.id, param.name
                )
            except ParameterPluginException as e:
                raise Arcor2Exception(f"Parameter {param.name} of action {action.name} has invalid value. {str(e)}")


def check_flows(
    parent: Union[CachedProject, common.ProjectFunction], action: common.Action, action_meta: ObjectAction
) -> None:
    """Raises exception if there is something wrong with flow(s).

    :param parent:
    :param action:
    :param action_meta:
    :return:
    """

    flow = action.flow()  # searches default flow (just this flow is supported so far)

    # it is ok to not specify any output (if the values are not going to be used anywhere)
    # return value(s) won't be stored in variable(s)
    if not flow.outputs:
        return

    # otherwise, all return values have to be stored in variables
    if len(flow.outputs) != len(action_meta.returns):
        raise Arcor2Exception("Number of the flow outputs does not match the number of action outputs.")

    for output in flow.outputs:
        hlp.is_valid_identifier(output)

    outputs: Set[str] = set()

    for act in parent.actions:
        for fl in act.flows:
            for output in fl.outputs:
                if output in outputs:
                    raise Arcor2Exception(f"Output '{output}' is not unique.")


def find_object_action(scene: CachedScene, action: common.Action) -> ObjectAction:

    obj_id, action_type = action.parse_type()
    obj = scene.object(obj_id)

    try:
        obj_type = glob.OBJECT_TYPES[obj.type]
    except KeyError:
        raise Arcor2Exception("Unknown object type.")

    try:
        act = obj_type.actions[action_type]
    except KeyError:
        raise Arcor2Exception("Unknown type of action.")

    if act.disabled:
        raise Arcor2Exception("Action is disabled.")

    return act


async def project_names() -> Set[str]:
    return {proj.name for proj in (await storage.get_projects()).items}


async def associated_projects(scene_id: str) -> Set[str]:
    return {project.id async for project in projects(scene_id)}


async def remove_object_references_from_projects(obj_id: str) -> None:

    assert glob.SCENE

    updated_project_ids: Set[str] = set()

    async for project in projects_using_object_as_parent(glob.SCENE.id, obj_id):

        # action_ids: Set[str] = set()

        # delete actions using the object
        for action_to_delete in {act.id for act in project.actions if act.parse_type()[0] == obj_id}:
            project.remove_action(action_to_delete)

        # delete actions using obj's action points as parameters
        # TODO fix this!
        """
        actions_using_invalid_param: Set[str] = \
            {act.id for act in ap.actions for param in act.parameters
             if param.type in (ActionParameterTypeEnum.JOINTS, ActionParameterTypeEnum.POSE) and
             param.value.startswith(obj_id)}

        ap.actions = [act for act in ap.actions if act.id not in actions_using_invalid_param]

        # get IDs of remaining actions
        action_ids.update({act.id for act in ap.actions})
        """

        # valid_ids: Set[str] = action_ids | ActionIOEnum.set()

        # TODO remove invalid logic items

        await storage.update_project(project.project)
        updated_project_ids.add(project.id)

    glob.logger.info("Updated projects: {}".format(updated_project_ids))


async def projects(scene_id: str) -> AsyncIterator[UpdateableCachedProject]:

    id_list = await storage.get_projects()

    for project_meta in id_list.items:

        project = await storage.get_project(project_meta.id)

        if project.scene_id != scene_id:
            continue

        yield UpdateableCachedProject(project)


def _project_using_object_as_parent(project: CachedProject, obj_id: str) -> bool:

    for ap in project.action_points:
        if ap.parent == obj_id:
            return True

    return False


def _project_referencing_object(project: CachedProject, obj_id: str) -> bool:

    for action in project.actions:
        action_obj_id, _ = action.parse_type()

        if action_obj_id == obj_id:
            return True

    return False


async def projects_using_object(scene_id: str, obj_id: str) -> AsyncIterator[UpdateableCachedProject]:
    """Combines functionality of projects_using_object_as_parent and
    projects_referencing_object.

    :param scene_id:
    :param obj_id:
    :return:
    """

    async for project in projects(scene_id):
        if _project_using_object_as_parent(project, obj_id) or _project_referencing_object(project, obj_id):
            yield project


async def projects_using_object_as_parent(scene_id: str, obj_id: str) -> AsyncIterator[UpdateableCachedProject]:

    async for project in projects(scene_id):
        if _project_using_object_as_parent(project, obj_id):
            yield project


async def invalidate_joints_using_object_as_parent(obj: common.SceneObject) -> None:

    assert glob.SCENE

    # Invalidates robot joints if action point's parent has changed its pose.
    async for project in projects_using_object_as_parent(glob.SCENE.id, obj.id):

        for ap in project.action_points:

            if ap.parent != obj.id:
                continue

            glob.logger.debug(f"Invalidating joints for {project.name}/{ap.name}.")
            project.invalidate_joints(ap.id)

        await storage.update_project(project.project)


async def projects_referencing_object(scene_id: str, obj_id: str) -> AsyncIterator[CachedProject]:

    async for project in projects(scene_id):
        if _project_referencing_object(project, obj_id):
            yield project


def project_problems(scene: CachedScene, project: CachedProject) -> List[str]:

    scene_objects: Dict[str, str] = {obj.id: obj.type for obj in scene.objects}

    action_ids: Set[str] = set()
    problems: List[str] = []

    unknown_types = {obj.type for obj in scene.objects} - glob.OBJECT_TYPES.keys()

    if unknown_types:
        return [f"Scene invalid, contains unknown types: {unknown_types}."]

    for ap in project.action_points:

        # test if all objects exists in scene
        if ap.parent and ap.parent not in scene_objects:
            problems.append(f"Action point '{ap.name}' has parent '{ap.parent}' that does not exist in the scene.")
            continue

        for joints in project.ap_joints(ap.id):
            if not joints.is_valid:
                problems.append(
                    f"Action point {ap.name} has invalid joints: {joints.name} " f"(robot {joints.robot_id})."
                )

        for action in project.actions:

            if action.id in action_ids:
                problems.append(f"Action {action.name} of the {ap.name} is not unique.")

            # check if objects have used actions
            obj_id, action_type = action.parse_type()

            if obj_id not in scene_objects.keys():
                problems.append(f"Object ID {obj_id} which action is used in {action.name} does not exist in scene.")
                continue

            try:
                os_type = scene_objects[obj_id]  # object type
            except KeyError:
                os_type = obj_id  # service

            if action_type not in glob.OBJECT_TYPES[os_type].actions:
                problems.append(
                    f"Object type {scene_objects[obj_id]} does not have action {action_type} " f"used in {action.id}."
                )
                continue

            try:
                check_action_params(scene, project, action, glob.OBJECT_TYPES[os_type].actions[action_type])
            except Arcor2Exception as e:
                problems.append(str(e))

    return problems


async def open_project(project_id: str) -> None:

    project = UpdateableCachedProject(await storage.get_project(project_id))

    if glob.SCENE:
        if glob.SCENE.id != project.scene_id:
            raise Arcor2Exception("Required project is associated to another scene.")
    else:
        await open_scene(project.scene_id)

    assert glob.SCENE
    for ap in project.action_points_with_parent:

        assert ap.parent

        if ap.parent not in glob.SCENE.object_ids | project.action_points_ids:
            glob.SCENE = None
            raise Arcor2Exception(f"Action point's {ap.name} parent not available.")

    glob.PROJECT = project



import asyncio
import base64
import os
import tempfile
import uuid
from typing import TYPE_CHECKING, Dict, Optional

import websockets
from websockets.server import WebSocketServerProtocol as WsClient

from arcor2 import helpers as hlp
from arcor2 import rest
from arcor2.data import common, rpc
from arcor2.exceptions import Arcor2Exception
from arcor2_arserver import events as server_events
from arcor2_arserver import globals as glob
from arcor2_arserver import notifications as notif
from arcor2_arserver import project
from arcor2_arserver.scene import scene_started, start_scene, stop_scene
from arcor2_arserver_data import events as sevts
from arcor2_build_data import URL as BUILD_URL
from arcor2_execution_data import URL as EXE_URL
from arcor2_execution_data import rpc as erpc

if TYPE_CHECKING:
    ReqQueue = asyncio.Queue[rpc.common.RPC.Request]
    RespQueue = asyncio.Queue[rpc.common.RPC.Response]
else:
    ReqQueue = asyncio.Queue
    RespQueue = asyncio.Queue

MANAGER_RPC_REQUEST_QUEUE: ReqQueue = ReqQueue()
MANAGER_RPC_RESPONSES: Dict[int, RespQueue] = {}


async def run_temp_package(package_id: str) -> None:

    assert glob.PROJECT
    project_id = glob.PROJECT.id
    glob.TEMPORARY_PACKAGE = True

    scene_online = scene_started()

    if scene_online:
        await stop_scene()  # the package will start it on its own

    await project.close_project()
    req = erpc.RunPackage.Request
    exe_req = req(uuid.uuid4().int, args=req.Args(package_id, cleanup_after_run=False))
    exe_resp = await manager_request(exe_req)

    if not exe_resp.result:
        glob.logger.warning(f"Execution of temporary package failed with: {exe_resp.messages}.")
    else:
        await server_events.package_started.wait()
        await server_events.package_stopped.wait()
        glob.logger.info("Temporary package stopped, let's remove it and reopen project.")

    glob.TEMPORARY_PACKAGE = False

    await manager_request(erpc.DeletePackage.Request(uuid.uuid4().int, args=rpc.common.IdArgs(package_id)))

    await project.open_project(project_id)

    assert glob.SCENE
    assert glob.PROJECT

    await notif.broadcast_event(sevts.p.OpenProject(sevts.p.OpenProject.Data(glob.SCENE.scene, glob.PROJECT.project)))

    if scene_online:
        await start_scene()


async def build_and_upload_package(project_id: str, package_name: str) -> str:
    """Builds package and uploads it to the Execution unit.

    :param project_id:
    :param package_name:
    :return: generated package ID.
    """

    package_id = common.uid("pkg")

    # call build service
    # TODO store data in memory
    with tempfile.TemporaryDirectory() as tmpdirname:
        path = os.path.join(tmpdirname, "publish.zip")

        await hlp.run_in_executor(
            rest.download,
            f"{BUILD_URL}/project/{project_id}/publish",
            path,
            {"packageName": package_name},
        )

        with open(path, "rb") as zip_file:
            b64_bytes = base64.b64encode(zip_file.read())
            b64_str = b64_bytes.decode()

    # send data to execution service
    exe_req = erpc.UploadPackage.Request(uuid.uuid4().int, args=erpc.UploadPackage.Request.Args(package_id, b64_str))
    exe_resp = await manager_request(exe_req)

    if not exe_resp.result:
        if not exe_resp.messages:
            raise Arcor2Exception("Upload to the Execution unit failed.")
        raise Arcor2Exception("\n".join(exe_resp.messages))

    return package_id


async def manager_request(req: rpc.common.RPC.Request, ui: Optional[WsClient] = None) -> rpc.common.RPC.Response:

    assert req.id not in MANAGER_RPC_RESPONSES

    MANAGER_RPC_RESPONSES[req.id] = RespQueue(maxsize=1)
    await MANAGER_RPC_REQUEST_QUEUE.put(req)
    resp = await MANAGER_RPC_RESPONSES[req.id].get()
    del MANAGER_RPC_RESPONSES[req.id]
    return resp


async def project_manager_client(handle_manager_incoming_messages) -> None:

    while True:

        glob.logger.info("Attempting connection to manager...")

        try:

            async with websockets.connect(EXE_URL) as manager_client:

                glob.logger.info("Connected to manager.")

                future = asyncio.ensure_future(handle_manager_incoming_messages(manager_client))

                while True:

                    if future.done():
                        break

                    try:
                        msg = await asyncio.wait_for(MANAGER_RPC_REQUEST_QUEUE.get(), 1.0)
                    except asyncio.TimeoutError:
                        continue

                    try:
                        await manager_client.send(msg.to_json())
                    except websockets.exceptions.ConnectionClosed:
                        await MANAGER_RPC_REQUEST_QUEUE.put(msg)
                        break
        except ConnectionRefusedError as e:
            glob.logger.error(e)
            await asyncio.sleep(delay=1.0)
